# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest
from unittest.mock import MagicMock

from libcloud.test import MockHttp
from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_ZONOMI
from libcloud.dns.drivers.zonomi import ZonomiDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures


class ZonomiTests(unittest.TestCase):
    def setUp(self):
        ZonomiDNSDriver.connectionCls.conn_class = ZonomiMockHttp
        ZonomiMockHttp.type = None
        self.driver = ZonomiDNSDriver(*DNS_PARAMS_ZONOMI)
        self.test_zone = Zone(
            id="zone.com",
            domain="zone.com",
            driver=self.driver,
            type="master",
            ttl=None,
            extra={},
        )
        self.test_record = Record(
            id="record.zone.com",
            name="record.zone.com",
            data="127.0.0.1",
            type="A",
            zone=self.test_zone,
            driver=self,
            extra={},
        )

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 3)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.TXT in record_types)

    def test_list_zones_empty(self):
        ZonomiMockHttp.type = "EMPTY_ZONES_LIST"
        zones = self.driver.list_zones()

        self.assertEqual(zones, [])

    def test_list_zones_success(self):
        zones = self.driver.list_zones()

        self.assertEqual(len(zones), 3)

        zone = zones[0]
        self.assertEqual(zone.id, "thegamertest.com")
        self.assertEqual(zone.domain, "thegamertest.com")
        self.assertEqual(zone.type, "master")
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

        second_zone = zones[1]
        self.assertEqual(second_zone.id, "lonelygamer.com")
        self.assertEqual(second_zone.domain, "lonelygamer.com")
        self.assertEqual(second_zone.type, "master")
        self.assertIsNone(second_zone.ttl)
        self.assertEqual(second_zone.driver, self.driver)

        third_zone = zones[2]
        self.assertEqual(third_zone.id, "gamertest.com")
        self.assertEqual(third_zone.domain, "gamertest.com")
        self.assertEqual(third_zone.type, "master")
        self.assertIsNone(third_zone.ttl)
        self.assertEqual(third_zone.driver, self.driver)

    def test_get_zone_GET_ZONE_DOES_NOT_EXIST(self):
        ZonomiMockHttp.type = "GET_ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_zone("testzone.com")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "testzone.com")
        else:
            self.fail("Exception was not thrown.")

    def test_get_zone_GET_ZONE_SUCCESS(self):
        ZonomiMockHttp.type = "GET_ZONE_SUCCESS"
        zone = self.driver.get_zone(zone_id="gamertest.com")

        self.assertEqual(zone.id, "gamertest.com")
        self.assertEqual(zone.domain, "gamertest.com")
        self.assertEqual(zone.type, "master")
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

    def test_delete_zone_DELETE_ZONE_DOES_NOT_EXIST(self):
        ZonomiMockHttp.type = "DELETE_ZONE_DOES_NOT_EXIST"
        try:
            self.driver.delete_zone(zone=self.test_zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, self.test_zone.id)
        else:
            self.fail("Exception was not thrown.")

    def test_delete_zone_delete_zone_success(self):
        ZonomiMockHttp.type = "DELETE_ZONE_SUCCESS"
        status = self.driver.delete_zone(zone=self.test_zone)

        self.assertEqual(status, True)

    def test_create_zone_already_exists(self):
        ZonomiMockHttp.type = "CREATE_ZONE_ALREADY_EXISTS"
        try:
            self.driver.create_zone(domain="gamertest.com")
        except ZoneAlreadyExistsError as e:
            self.assertEqual(e.zone_id, "gamertest.com")
        else:
            self.fail("Exception was not thrown.")

    def test_create_zone_create_zone_success(self):
        ZonomiMockHttp.type = "CREATE_ZONE_SUCCESS"

        zone = self.driver.create_zone(domain="myzone.com")

        self.assertEqual(zone.id, "myzone.com")
        self.assertEqual(zone.domain, "myzone.com")
        self.assertEqual(zone.type, "master")
        self.assertIsNone(zone.ttl)

    def test_list_records_empty_list(self):
        ZonomiMockHttp.type = "LIST_RECORDS_EMPTY_LIST"
        pass

    def test_list_records_success(self):
        ZonomiMockHttp.type = "LIST_RECORDS_SUCCESS"
        records = self.driver.list_records(zone=self.test_zone)

        self.assertEqual(len(records), 4)

        record = records[0]
        self.assertEqual(record.id, "zone.com")
        self.assertEqual(record.type, "SOA")
        self.assertEqual(record.data, "ns1.zonomi.com. soacontact.zonomi.com. 13")
        self.assertEqual(record.name, "zone.com")
        self.assertEqual(record.zone, self.test_zone)

        second_record = records[1]
        self.assertEqual(second_record.id, "zone.com")
        self.assertEqual(second_record.name, "zone.com")
        self.assertEqual(second_record.type, "NS")
        self.assertEqual(second_record.data, "ns1.zonomi.com")
        self.assertEqual(second_record.zone, self.test_zone)

        third_record = records[2]
        self.assertEqual(third_record.id, "oltjano")
        self.assertEqual(third_record.name, "oltjano")
        self.assertEqual(third_record.type, "A")
        self.assertEqual(third_record.data, "127.0.0.1")
        self.assertEqual(third_record.zone, self.test_zone)

        fourth_record = records[3]
        self.assertEqual(fourth_record.id, "zone.com")
        self.assertEqual(fourth_record.name, "zone.com")
        self.assertEqual(fourth_record.type, "NS")
        self.assertEqual(fourth_record.data, "ns5.zonomi.com")
        self.assertEqual(fourth_record.zone, self.test_zone)

    def test_get_record_does_not_exist(self):
        ZonomiMockHttp.type = "GET_RECORD_DOES_NOT_EXIST"
        zone = Zone(
            id="zone.com",
            domain="zone.com",
            type="master",
            ttl=None,
            driver=self.driver,
        )
        self.driver.get_zone = MagicMock(return_value=zone)
        record_id = "nonexistent"
        try:
            self.driver.get_record(record_id=record_id, zone_id="zone.com")
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record_id)
        else:
            self.fail("Exception was not thrown.")

    def test_get_record_success(self):
        ZonomiMockHttp.type = "GET_RECORD_SUCCESS"
        zone = Zone(
            id="zone.com",
            domain="zone.com",
            type="master",
            ttl=None,
            driver=self.driver,
        )
        self.driver.get_zone = MagicMock(return_value=zone)
        record = self.driver.get_record(record_id="oltjano", zone_id="zone.com")

        self.assertEqual(record.id, "oltjano")
        self.assertEqual(record.name, "oltjano")
        self.assertEqual(record.type, "A")
        self.assertEqual(record.data, "127.0.0.1")

    def test_delete_record_does_not_exist(self):
        ZonomiMockHttp.type = "DELETE_RECORD_DOES_NOT_EXIST"
        record = self.test_record
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail("Exception was not thrown.")

    def test_delete_record_success(self):
        ZonomiMockHttp.type = "DELETE_RECORD_SUCCESS"
        record = self.test_record
        status = self.driver.delete_record(record=record)

        self.assertEqual(status, True)

    def test_create_record_already_exists(self):
        zone = self.test_zone
        ZonomiMockHttp.type = "CREATE_RECORD_ALREADY_EXISTS"
        try:
            self.driver.create_record(
                name="createrecord", type="A", data="127.0.0.1", zone=zone, extra={}
            )
        except RecordAlreadyExistsError as e:
            self.assertEqual(e.record_id, "createrecord")
        else:
            self.fail("Exception was not thrown.")

    def test_create_record_success(self):
        ZonomiMockHttp.type = "CREATE_RECORD_SUCCESS"
        zone = self.test_zone
        record = self.driver.create_record(
            name="createrecord", zone=zone, type="A", data="127.0.0.1", extra={}
        )

        self.assertEqual(record.id, "createrecord")
        self.assertEqual(record.name, "createrecord")
        self.assertEqual(record.type, "A")
        self.assertEqual(record.data, "127.0.0.1")
        self.assertEqual(record.zone, zone)

    def test_convert_to_slave(self):
        zone = self.test_zone
        result = self.driver.ex_convert_to_secondary(zone, "1.2.3.4")
        self.assertTrue(result)

    def test_convert_to_slave_couldnt_convert(self):
        zone = self.test_zone
        ZonomiMockHttp.type = "COULDNT_CONVERT"
        try:
            self.driver.ex_convert_to_secondary(zone, "1.2.3.4")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "zone.com")
        else:
            self.fail("Exception was not thrown.")

    def test_convert_to_master(self):
        zone = self.test_zone
        result = self.driver.ex_convert_to_master(zone)
        self.assertTrue(result)

    def test_convert_to_master_couldnt_convert(self):
        zone = self.test_zone
        ZonomiMockHttp.type = "COULDNT_CONVERT"
        try:
            self.driver.ex_convert_to_master(zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "zone.com")
        else:
            self.fail("Exception was not thrown.")


class ZonomiMockHttp(MockHttp):
    fixtures = DNSFileFixtures("zonomi")

    def _app_dns_dyndns_jsp_EMPTY_ZONES_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_zones_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_GET_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_GET_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_DELETE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone_does_not_exist.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_DELETE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_addzone_jsp_CREATE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_addzone_jsp_CREATE_ZONE_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_already_exists.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_LIST_RECORDS_EMPTY_LIST(self, method, url, body, headers):
        body = self.fixtures.load("list_records_empty_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_LIST_RECORDS_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_DELETE_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_DELETE_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("delete_record_does_not_exist.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_CREATE_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("create_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_CREATE_RECORD_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("create_record_already_exists.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_GET_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_dyndns_jsp_GET_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_converttosecondary_jsp(self, method, url, body, headers):
        body = self.fixtures.load("converted_to_slave.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_converttosecondary_jsp_COULDNT_CONVERT(self, method, url, body, headers):
        body = self.fixtures.load("couldnt_convert.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_converttomaster_jsp(self, method, url, body, headers):
        body = self.fixtures.load("converted_to_master.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _app_dns_converttomaster_jsp_COULDNT_CONVERT(self, method, url, body, headers):
        body = self.fixtures.load("couldnt_convert.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import hmac
import base64
import hashlib

from libcloud.utils.py3 import b, httplib, urlquote, urlencode
from libcloud.common.base import JsonResponse, PollingConnection, ConnectionUserAndKey
from libcloud.common.types import ProviderError, MalformedResponseError
from libcloud.compute.types import InvalidCredsError


class CloudStackResponse(JsonResponse):
    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError("Invalid provider credentials")

        value = None
        body = self.parse_body()
        if hasattr(body, "values"):
            values = list(body.values())[0]
            if "errortext" in values:
                value = values["errortext"]
        if value is None:
            value = self.body

        if not value:
            value = "WARNING: error message text sent by provider was empty."

        error = ProviderError(value=value, http_code=self.status, driver=self.connection.driver)
        raise error


class CloudStackConnection(ConnectionUserAndKey, PollingConnection):
    responseCls = CloudStackResponse
    poll_interval = 1
    request_method = "_sync_request"
    timeout = 600

    ASYNC_PENDING = 0
    ASYNC_SUCCESS = 1
    ASYNC_FAILURE = 2

    def encode_data(self, data):
        """
        Must of the data is sent as part of query params (eeww),
        but in newer versions, userdata argument can be sent as a
        urlencoded data in the request body.
        """
        if data:
            data = urlencode(data)

        return data

    def _make_signature(self, params):
        signature = [(k.lower(), v) for k, v in list(params.items())]
        signature.sort(key=lambda x: x[0])

        pairs = []
        for pair in signature:
            key = urlquote(str(pair[0]), safe="[]")
            value = urlquote(str(pair[1]), safe="[]*")
            item = "{}={}".format(key, value)
            pairs.append(item)

        signature = "&".join(pairs)

        signature = signature.lower().replace("+", "%20")
        signature = hmac.new(b(self.key), msg=b(signature), digestmod=hashlib.sha1)
        return base64.b64encode(b(signature.digest()))

    def add_default_params(self, params):
        params["apiKey"] = self.user_id
        params["response"] = "json"

        return params

    def pre_connect_hook(self, params, headers):
        params["signature"] = self._make_signature(params)

        return params, headers

    def _async_request(
        self,
        command,
        action=None,
        params=None,
        data=None,
        headers=None,
        method="GET",
        context=None,
    ):
        if params:
            context = copy.deepcopy(params)
        else:
            context = {}

        # Command is specified as part of GET call
        context["command"] = command
        result = super().async_request(
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            context=context,
        )
        return result["jobresult"]

    def get_request_kwargs(
        self, action, params=None, data="", headers=None, method="GET", context=None
    ):
        command = context["command"]
        request_kwargs = {
            "command": command,
            "action": action,
            "params": params,
            "data": data,
            "headers": headers,
            "method": method,
        }
        return request_kwargs

    def get_poll_request_kwargs(self, response, context, request_kwargs):
        job_id = response["jobid"]
        params = {"jobid": job_id}
        kwargs = {"command": "queryAsyncJobResult", "params": params}
        return kwargs

    def has_completed(self, response):
        status = response.get("jobstatus", self.ASYNC_PENDING)

        if status == self.ASYNC_FAILURE:
            msg = response.get("jobresult", {}).get("errortext", status)
            raise Exception(msg)

        return status == self.ASYNC_SUCCESS

    def _sync_request(
        self, command, action=None, params=None, data=None, headers=None, method="GET"
    ):
        """
        This method handles synchronous calls which are generally fast
        information retrieval requests and thus return 'quickly'.
        """
        # command is always sent as part of "command" query parameter
        if params:
            params = copy.deepcopy(params)
        else:
            params = {}

        params["command"] = command
        # pylint: disable=maybe-no-member
        result = self.request(
            action=self.driver.path,
            params=params,
            data=data,
            headers=headers,
            method=method,
        )

        command = command.lower()

        # Work around for older versions which don't return "response" suffix
        # in delete ingress rule response command name
        if (
            command == "revokesecuritygroupingress"
            and "revokesecuritygroupingressresponse" not in result.object
        ):
            command = command
        elif command == "restorevirtualmachine" and "restorevmresponse" in result.object:
            command = "restorevmresponse"
        else:
            command = command + "response"

        if command not in result.object:
            raise MalformedResponseError(
                "Unknown response format {}".format(command),
                body=result.body,
                driver=self.driver,
            )
        result = result.object[command]
        return result


class CloudStackDriverMixIn:
    host = None
    path = None

    connectionCls = CloudStackConnection

    def __init__(self, key, secret=None, secure=True, host=None, port=None):
        host = host or self.host
        super().__init__(key, secret, secure, host, port)

    def _sync_request(
        self, command, action=None, params=None, data=None, headers=None, method="GET"
    ):
        return self.connection._sync_request(  # pylint: disable=maybe-no-member
            command=command,
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
        )

    def _async_request(
        self,
        command,
        action=None,
        params=None,
        data=None,
        headers=None,
        method="GET",
        context=None,
    ):
        return self.connection._async_request(  # pylint: disable=maybe-no-member
            command=command,
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            context=context,
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["BaseRangeDownloadMockHttp"]

from typing import Tuple

from libcloud.test import MockHttp  # pylint: disable-msg=E0611


class BaseRangeDownloadMockHttp(MockHttp):
    """
    Base MockHttp class which implements some utility methods for asserting
    range download requests.
    """

    def _get_start_and_end_bytes_from_range_str(self, range_str, body):
        # type: (str, str) -> Tuple[int, int]
        range_str = range_str.split("bytes=")[1]
        range_str = range_str.split("-")
        range_str = [value for value in range_str if value.strip()]
        start_bytes = int(range_str[0])

        if len(range_str) == 2:
            end_bytes = int(range_str[1])
        else:
            end_bytes = len(body)

        return start_bytes, end_bytes



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.common.types import InvalidCredsError
from libcloud.test.secrets import DIGITALOCEAN_v2_PARAMS
from libcloud.test.file_fixtures import FileFixtures
from libcloud.common.digitalocean import DigitalOceanBaseDriver


class DigitalOceanTests(LibcloudTestCase):
    def setUp(self):
        DigitalOceanBaseDriver.connectionCls.conn_class = DigitalOceanMockHttp
        DigitalOceanMockHttp.type = None
        self.driver = DigitalOceanBaseDriver(*DIGITALOCEAN_v2_PARAMS)

    def test_authentication(self):
        DigitalOceanMockHttp.type = "UNAUTHORIZED"
        self.assertRaises(InvalidCredsError, self.driver.ex_account_info)

    def test_ex_account_info(self):
        account_info = self.driver.ex_account_info()
        self.assertEqual(account_info["uuid"], "a1234567890b1234567890c1234567890d12345")
        self.assertTrue(account_info["email_verified"])
        self.assertEqual(account_info["email"], "user@domain.tld")
        self.assertEqual(account_info["droplet_limit"], 10)

    def test_ex_list_events(self):
        events = self.driver.ex_list_events()
        self.assertEqual(events, [])

    def test_ex_get_event(self):
        action = self.driver.ex_get_event("12345670")
        self.assertEqual(action["id"], 12345670)
        self.assertEqual(action["status"], "completed")
        self.assertEqual(action["type"], "power_on")

    def test__paginated_request(self):
        DigitalOceanMockHttp.type = "page_1"
        actions = self.driver._paginated_request("/v2/actions", "actions")
        self.assertEqual(actions[0]["id"], 12345671)
        self.assertEqual(actions[0]["status"], "completed")


class DigitalOceanMockHttp(MockHttp):
    fixtures = FileFixtures("common", "digitalocean")

    response = {
        None: httplib.OK,
        "CREATE": httplib.CREATED,
        "DELETE": httplib.NO_CONTENT,
        "EMPTY": httplib.OK,
        "NOT_FOUND": httplib.NOT_FOUND,
        "UNAUTHORIZED": httplib.UNAUTHORIZED,
        "UPDATE": httplib.OK,
    }

    def _v2_account(self, method, url, body, headers):
        body = self.fixtures.load("_v2_account.json")
        return (
            self.response[self.type],
            body,
            {},
            httplib.responses[self.response[self.type]],
        )

    def _v2_account_UNAUTHORIZED(self, method, url, body, headers):
        body = self.fixtures.load("_v2_account_UNAUTHORIZED.json")
        return (
            self.response[self.type],
            body,
            {},
            httplib.responses[self.response[self.type]],
        )

    def _v2_actions(self, method, url, body, headers):
        body = self.fixtures.load("_v2_actions.json")
        return (
            self.response[self.type],
            body,
            {},
            httplib.responses[self.response[self.type]],
        )

    def _v2_actions_12345670(self, method, url, body, headers):
        body = self.fixtures.load("_v2_actions_12345670.json")
        return (
            self.response[self.type],
            body,
            {},
            httplib.responses[self.response[self.type]],
        )

    def _v2_actions_page_1(self, method, url, body, headers):
        body = self.fixtures.load("_v2_actions_page_1.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance withv
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.utils.py3 import ET
from libcloud.utils.xml import findall, findtext, fixxpath
from libcloud.utils.misc import reverse_dict
from libcloud.loadbalancer.base import Driver, Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State, Provider
from libcloud.common.dimensiondata import (
    TYPES_URN,
    API_ENDPOINTS,
    DEFAULT_REGION,
    DimensionDataPool,
    DimensionDataVIPNode,
    DimensionDataConnection,
    DimensionDataPoolMember,
    DimensionDataDefaultiRule,
    DimensionDataVirtualListener,
    DimensionDataPersistenceProfile,
    DimensionDataDefaultHealthMonitor,
    DimensionDataVirtualListenerCompatibility,
)


class DimensionDataLBDriver(Driver):
    """
    DimensionData node driver.
    """

    selected_region = None
    connectionCls = DimensionDataConnection
    name = "Dimension Data Load Balancer"
    website = "https://cloud.dimensiondata.com/"
    type = Provider.DIMENSIONDATA
    api_version = 1.0

    network_domain_id = None

    _VALUE_TO_ALGORITHM_MAP = {
        "ROUND_ROBIN": Algorithm.ROUND_ROBIN,
        "LEAST_CONNECTIONS": Algorithm.LEAST_CONNECTIONS,
        "SHORTEST_RESPONSE": Algorithm.SHORTEST_RESPONSE,
        "PERSISTENT_IP": Algorithm.PERSISTENT_IP,
    }
    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    _VALUE_TO_STATE_MAP = {
        "NORMAL": State.RUNNING,
        "PENDING_ADD": State.PENDING,
        "PENDING_CHANGE": State.PENDING,
        "PENDING_DELETE": State.PENDING,
        "FAILED_ADD": State.ERROR,
        "FAILED_CHANGE": State.ERROR,
        "FAILED_DELETE": State.ERROR,
        "REQUIRES_SUPPORT": State.ERROR,
    }

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=None,
        region=DEFAULT_REGION,
        **kwargs,
    ):
        if region not in API_ENDPOINTS and host is None:
            raise ValueError("Invalid region: %s, no host specified" % (region))
        if region is not None:
            self.selected_region = API_ENDPOINTS[region]

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            api_version=api_version,
            region=region,
            **kwargs,
        )

    def _ex_connection_class_kwargs(self):
        """
        Add the region to the kwargs before the connection is instantiated
        """

        kwargs = super()._ex_connection_class_kwargs()
        kwargs["region"] = self.selected_region
        return kwargs

    def create_balancer(
        self,
        name,
        port=None,
        protocol=None,
        algorithm=None,
        members=None,
        ex_listener_ip_address=None,
    ):
        """
        Create a new load balancer instance

        :param name: Name of the new load balancer (required)
        :type  name: ``str``

        :param port: An integer in the range of 1-65535. If not supplied,
                     it will be taken to mean 'Any Port'
        :type  port: ``int``

        :param protocol: Loadbalancer protocol, defaults to http.
        :type  protocol: ``str``

        :param members: list of Members to attach to balancer (optional)
        :type  members: ``list`` of :class:`Member`

        :param algorithm: Load balancing algorithm, defaults to ROUND_ROBIN.
        :type algorithm: :class:`.Algorithm`

        :param ex_listener_ip_address: Must be a valid IPv4 in dot-decimal
                                       notation (x.x.x.x).
        :type ex_listener_ip_address: ``str``

        :rtype: :class:`LoadBalancer`
        """
        network_domain_id = self.network_domain_id
        if protocol is None:
            protocol = "http"
        if algorithm is None:
            algorithm = Algorithm.ROUND_ROBIN

        # Create a pool first
        pool = self.ex_create_pool(
            network_domain_id=network_domain_id,
            name=name,
            ex_description=None,
            balancer_method=self._ALGORITHM_TO_VALUE_MAP[algorithm],
        )

        # Attach the members to the pool as nodes
        if members is not None:
            for member in members:
                node = self.ex_create_node(
                    network_domain_id=network_domain_id,
                    name=member.ip,
                    ip=member.ip,
                    ex_description=None,
                )
                self.ex_create_pool_member(pool=pool, node=node, port=port)

        # Create the virtual listener (balancer)
        listener = self.ex_create_virtual_listener(
            network_domain_id=network_domain_id,
            name=name,
            ex_description=name,
            port=port,
            pool=pool,
            listener_ip_address=ex_listener_ip_address,
        )

        return LoadBalancer(
            id=listener.id,
            name=listener.name,
            state=State.RUNNING,
            ip=listener.ip,
            port=port,
            driver=self,
            extra={
                "pool_id": pool.id,
                "network_domain_id": network_domain_id,
                "listener_ip_address": ex_listener_ip_address,
            },
        )

    def list_balancers(self, ex_network_domain_id=None):
        """
        List all loadbalancers inside a geography or in given network.

        In Dimension Data terminology these are known as virtual listeners

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :rtype: ``list`` of :class:`LoadBalancer`
        """
        params = None
        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        return self._to_balancers(
            self.connection.request_with_orgId_api_2(
                "networkDomainVip/virtualListener", params=params
            ).object
        )

    def get_balancer(self, balancer_id):
        """
        Return a :class:`LoadBalancer` object.

        :param balancer_id: id of a load balancer you want to fetch
        :type  balancer_id: ``str``

        :rtype: :class:`LoadBalancer`
        """

        bal = self.connection.request_with_orgId_api_2(
            "networkDomainVip/virtualListener/%s" % balancer_id
        ).object
        return self._to_balancer(bal)

    def list_protocols(self):
        """
        Return a list of supported protocols.

        Since all protocols are support by Dimension Data, this is a list
        of common protocols.

        :rtype: ``list`` of ``str``
        """
        return ["http", "https", "tcp", "udp", "ftp", "smtp"]

    def balancer_list_members(self, balancer):
        """
        Return list of members attached to balancer.

        In Dimension Data terminology these are the members of the pools
        within a virtual listener.

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`Member`
        """
        pool_members = self.ex_get_pool_members(balancer.extra["pool_id"])
        members = []
        for pool_member in pool_members:
            members.append(
                Member(
                    id=pool_member.id,
                    ip=pool_member.ip,
                    port=pool_member.port,
                    balancer=balancer,
                    extra=None,
                )
            )
        return members

    def balancer_attach_member(self, balancer, member):
        """
        Attach a member to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member to join to the balancer
        :type member: :class:`Member`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """
        node = self.ex_create_node(
            network_domain_id=balancer.extra["network_domain_id"],
            name="Member." + member.ip,
            ip=member.ip,
            ex_description="",
        )
        if node is False:
            return False
        pool = self.ex_get_pool(balancer.extra["pool_id"])
        pool_member = self.ex_create_pool_member(pool=pool, node=node, port=member.port)
        member.id = pool_member.id
        return member

    def balancer_detach_member(self, balancer, member):
        """
        Detach member from balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :return: ``True`` if member detach was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        create_pool_m = ET.Element("removePoolMember", {"xmlns": TYPES_URN, "id": member.id})

        result = self.connection.request_with_orgId_api_2(
            "networkDomainVip/removePoolMember",
            method="POST",
            data=ET.tostring(create_pool_m),
        ).object
        response_code = findtext(result, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def destroy_balancer(self, balancer):
        """
        Destroy a load balancer (virtual listener)

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :return: ``True`` if the destroy was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        delete_listener = ET.Element(
            "deleteVirtualListener", {"xmlns": TYPES_URN, "id": balancer.id}
        )

        result = self.connection.request_with_orgId_api_2(
            "networkDomainVip/deleteVirtualListener",
            method="POST",
            data=ET.tostring(delete_listener),
        ).object
        response_code = findtext(result, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def ex_set_current_network_domain(self, network_domain_id):
        """
        Set the network domain (part of the network) of the driver

        :param network_domain_id: ID of the pool (required)
        :type  network_domain_id: ``str``
        """
        self.network_domain_id = network_domain_id

    def ex_get_current_network_domain(self):
        """
        Get the current network domain ID of the driver.

        :return: ID of the network domain
        :rtype: ``str``
        """
        return self.network_domain_id

    def ex_create_pool_member(self, pool, node, port=None):
        """
        Create a new member in an existing pool from an existing node

        :param pool: Instance of ``DimensionDataPool`` (required)
        :type  pool: ``DimensionDataPool``

        :param node: Instance of ``DimensionDataVIPNode`` (required)
        :type  node: ``DimensionDataVIPNode``

        :param port: Port the the service will listen on
        :type  port: ``str``

        :return: The node member, instance of ``DimensionDataPoolMember``
        :rtype: ``DimensionDataPoolMember``
        """
        create_pool_m = ET.Element("addPoolMember", {"xmlns": TYPES_URN})
        ET.SubElement(create_pool_m, "poolId").text = pool.id
        ET.SubElement(create_pool_m, "nodeId").text = node.id
        if port is not None:
            ET.SubElement(create_pool_m, "port").text = str(port)
        ET.SubElement(create_pool_m, "status").text = "ENABLED"

        response = self.connection.request_with_orgId_api_2(
            "networkDomainVip/addPoolMember",
            method="POST",
            data=ET.tostring(create_pool_m),
        ).object

        member_id = None
        node_name = None
        for info in findall(response, "info", TYPES_URN):
            if info.get("name") == "poolMemberId":
                member_id = info.get("value")
            if info.get("name") == "nodeName":
                node_name = info.get("value")

        return DimensionDataPoolMember(
            id=member_id,
            name=node_name,
            status=State.RUNNING,
            ip=node.ip,
            port=port,
            node_id=node.id,
        )

    def ex_create_node(
        self,
        network_domain_id,
        name,
        ip,
        ex_description,
        connection_limit=25000,
        connection_rate_limit=2000,
    ):
        """
        Create a new node

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the node (required)
        :type  name: ``str``

        :param ip: IPv4 address of the node (required)
        :type  ip: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param connection_limit: Maximum number
                of concurrent connections per sec
        :type  connection_limit: ``int``

        :param connection_rate_limit: Maximum number of concurrent sessions
        :type  connection_rate_limit: ``int``

        :return: Instance of ``DimensionDataVIPNode``
        :rtype: ``DimensionDataVIPNode``
        """
        create_node_elm = ET.Element("createNode", {"xmlns": TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId").text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text = str(ex_description)
        ET.SubElement(create_node_elm, "ipv4Address").text = ip
        ET.SubElement(create_node_elm, "status").text = "ENABLED"
        ET.SubElement(create_node_elm, "connectionLimit").text = str(connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit").text = str(connection_rate_limit)

        response = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/createNode",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object

        node_id = None
        node_name = None
        for info in findall(response, "info", TYPES_URN):
            if info.get("name") == "nodeId":
                node_id = info.get("value")
            if info.get("name") == "name":
                node_name = info.get("value")
        return DimensionDataVIPNode(id=node_id, name=node_name, status=State.RUNNING, ip=ip)

    def ex_update_node(self, node):
        """
        Update the properties of a node

        :param pool: The instance of ``DimensionDataNode`` to update
        :type  pool: ``DimensionDataNode``

        :return: The instance of ``DimensionDataNode``
        :rtype: ``DimensionDataNode``
        """
        create_node_elm = ET.Element("editNode", {"xmlns": TYPES_URN})
        ET.SubElement(create_node_elm, "connectionLimit").text = str(node.connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit").text = str(node.connection_rate_limit)

        self.connection.request_with_orgId_api_2(
            action="networkDomainVip/createNode",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object
        return node

    def ex_set_node_state(self, node, enabled):
        """
        Change the state of a node (enable/disable)

        :param pool: The instance of ``DimensionDataNode`` to update
        :type  pool: ``DimensionDataNode``

        :param enabled: The target state of the node
        :type  enabled: ``bool``

        :return: The instance of ``DimensionDataNode``
        :rtype: ``DimensionDataNode``
        """
        create_node_elm = ET.Element("editNode", {"xmlns": TYPES_URN})
        ET.SubElement(create_node_elm, "status").text = "ENABLED" if enabled is True else "DISABLED"

        self.connection.request_with_orgId_api_2(
            action="networkDomainVip/editNode",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object
        return node

    def ex_create_pool(
        self,
        network_domain_id,
        name,
        balancer_method,
        ex_description,
        health_monitors=None,
        service_down_action="NONE",
        slow_ramp_time=30,
    ):
        """
        Create a new pool

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the node (required)
        :type  name: ``str``

        :param balancer_method: The load balancer algorithm (required)
        :type  balancer_method: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param health_monitors: A list of health monitors to use for the pool.
        :type  health_monitors: ``list`` of
            :class:`DimensionDataDefaultHealthMonitor`

        :param service_down_action: What to do when node
                                    is unavailable NONE, DROP or RESELECT
        :type  service_down_action: ``str``

        :param slow_ramp_time: Number of seconds to stagger ramp up of nodes
        :type  slow_ramp_time: ``int``

        :return: Instance of ``DimensionDataPool``
        :rtype: ``DimensionDataPool``
        """
        # Names cannot contain spaces.
        name.replace(" ", "_")
        create_node_elm = ET.Element("createPool", {"xmlns": TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId").text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text = str(ex_description)
        ET.SubElement(create_node_elm, "loadBalanceMethod").text = str(balancer_method)

        if health_monitors is not None:
            for monitor in health_monitors:
                ET.SubElement(create_node_elm, "healthMonitorId").text = str(monitor.id)

        ET.SubElement(create_node_elm, "serviceDownAction").text = service_down_action
        ET.SubElement(create_node_elm, "slowRampTime").text = str(slow_ramp_time)

        response = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/createPool",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object

        pool_id = None
        for info in findall(response, "info", TYPES_URN):
            if info.get("name") == "poolId":
                pool_id = info.get("value")

        return DimensionDataPool(
            id=pool_id,
            name=name,
            description=ex_description,
            status=State.RUNNING,
            load_balance_method=str(balancer_method),
            health_monitor_id=None,
            service_down_action=service_down_action,
            slow_ramp_time=str(slow_ramp_time),
        )

    def ex_create_virtual_listener(
        self,
        network_domain_id,
        name,
        ex_description,
        port=None,
        pool=None,
        listener_ip_address=None,
        persistence_profile=None,
        fallback_persistence_profile=None,
        irule=None,
        protocol="TCP",
        connection_limit=25000,
        connection_rate_limit=2000,
        source_port_preservation="PRESERVE",
    ):
        """
        Create a new virtual listener (load balancer)

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the listener (required)
        :type  name: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param port: An integer in the range of 1-65535. If not supplied,
                     it will be taken to mean 'Any Port'
        :type  port: ``int``

        :param pool: The pool to use for the listener
        :type  pool: :class:`DimensionDataPool`

        :param listener_ip_address: The IPv4 Address of the virtual listener
        :type  listener_ip_address: ``str``

        :param persistence_profile: Persistence profile
        :type  persistence_profile: :class:`DimensionDataPersistenceProfile`

        :param fallback_persistence_profile: Fallback persistence profile
        :type  fallback_persistence_profile:
            :class:`DimensionDataPersistenceProfile`

        :param irule: The iRule to apply
        :type  irule: :class:`DimensionDataDefaultiRule`

        :param protocol: For STANDARD type, ANY, TCP or UDP
                         for PERFORMANCE_LAYER_4 choice of ANY, TCP, UDP, HTTP
        :type  protocol: ``str``

        :param connection_limit: Maximum number
                                of concurrent connections per sec
        :type  connection_limit: ``int``

        :param connection_rate_limit: Maximum number of concurrent sessions
        :type  connection_rate_limit: ``int``

        :param source_port_preservation: Choice of PRESERVE,
                                        PRESERVE_STRICT or CHANGE
        :type  source_port_preservation: ``str``

        :return: Instance of the listener
        :rtype: ``DimensionDataVirtualListener``
        """
        if (port == 80) or (port == 443):
            listener_type = "PERFORMANCE_LAYER_4"
        else:
            listener_type = "STANDARD"

        create_node_elm = ET.Element("createVirtualListener", {"xmlns": TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId").text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text = str(ex_description)
        ET.SubElement(create_node_elm, "type").text = listener_type
        ET.SubElement(create_node_elm, "protocol").text = protocol
        if listener_ip_address is not None:
            ET.SubElement(create_node_elm, "listenerIpAddress").text = str(listener_ip_address)
        if port is not None:
            ET.SubElement(create_node_elm, "port").text = str(port)
        ET.SubElement(create_node_elm, "enabled").text = "true"
        ET.SubElement(create_node_elm, "connectionLimit").text = str(connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit").text = str(connection_rate_limit)
        ET.SubElement(create_node_elm, "sourcePortPreservation").text = source_port_preservation
        if pool is not None:
            ET.SubElement(create_node_elm, "poolId").text = pool.id
        if persistence_profile is not None:
            ET.SubElement(create_node_elm, "persistenceProfileId").text = persistence_profile.id
        if fallback_persistence_profile is not None:
            ET.SubElement(
                create_node_elm, "fallbackPersistenceProfileId"
            ).text = fallback_persistence_profile.id
        if irule is not None:
            ET.SubElement(create_node_elm, "iruleId").text = irule.id

        response = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/createVirtualListener",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object

        virtual_listener_id = None
        virtual_listener_ip = None
        for info in findall(response, "info", TYPES_URN):
            if info.get("name") == "virtualListenerId":
                virtual_listener_id = info.get("value")
            if info.get("name") == "listenerIpAddress":
                virtual_listener_ip = info.get("value")

        return DimensionDataVirtualListener(
            id=virtual_listener_id,
            name=name,
            ip=virtual_listener_ip,
            status=State.RUNNING,
        )

    def ex_get_pools(self, ex_network_domain_id=None):
        """
        Get all of the pools inside the current geography or
        in given network.

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :return: Returns a ``list`` of type ``DimensionDataPool``
        :rtype: ``list`` of ``DimensionDataPool``
        """
        params = None

        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        pools = self.connection.request_with_orgId_api_2(
            "networkDomainVip/pool", params=params
        ).object

        return self._to_pools(pools)

    def ex_get_pool(self, pool_id):
        """
        Get a specific pool inside the current geography

        :param pool_id: The identifier of the pool
        :type  pool_id: ``str``

        :return: Returns an instance of ``DimensionDataPool``
        :rtype: ``DimensionDataPool``
        """
        pool = self.connection.request_with_orgId_api_2("networkDomainVip/pool/%s" % pool_id).object
        return self._to_pool(pool)

    def ex_update_pool(self, pool):
        """
        Update the properties of an existing pool
        only method, serviceDownAction and slowRampTime are updated

        :param pool: The instance of ``DimensionDataPool`` to update
        :type  pool: ``DimensionDataPool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        create_node_elm = ET.Element("editPool", {"xmlns": TYPES_URN})

        ET.SubElement(create_node_elm, "loadBalanceMethod").text = str(pool.load_balance_method)
        ET.SubElement(create_node_elm, "serviceDownAction").text = pool.service_down_action
        ET.SubElement(create_node_elm, "slowRampTime").text = str(pool.slow_ramp_time)

        response = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/editPool",
            method="POST",
            data=ET.tostring(create_node_elm),
        ).object
        response_code = findtext(response, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def ex_destroy_pool(self, pool):
        """
        Destroy an existing pool

        :param pool: The instance of ``DimensionDataPool`` to destroy
        :type  pool: ``DimensionDataPool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        destroy_request = ET.Element("deletePool", {"xmlns": TYPES_URN, "id": pool.id})

        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/deletePool",
            method="POST",
            data=ET.tostring(destroy_request),
        ).object
        response_code = findtext(result, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def ex_get_pool_members(self, pool_id):
        """
        Get the members of a pool

        :param pool: The instance of a pool
        :type  pool: ``DimensionDataPool``

        :return: Returns an ``list`` of ``DimensionDataPoolMember``
        :rtype: ``list`` of ``DimensionDataPoolMember``
        """
        members = self.connection.request_with_orgId_api_2(
            "networkDomainVip/poolMember?poolId=%s" % pool_id
        ).object
        return self._to_members(members)

    def ex_get_pool_member(self, pool_member_id):
        """
        Get a specific member of a pool

        :param pool: The id of a pool member
        :type  pool: ``str``

        :return: Returns an instance of ``DimensionDataPoolMember``
        :rtype: ``DimensionDataPoolMember``
        """
        member = self.connection.request_with_orgId_api_2(
            "networkDomainVip/poolMember/%s" % pool_member_id
        ).object
        return self._to_member(member)

    def ex_set_pool_member_state(self, member, enabled=True):
        request = ET.Element("editPoolMember", {"xmlns": TYPES_URN, "id": member.id})
        state = "ENABLED" if enabled is True else "DISABLED"
        ET.SubElement(request, "status").text = state

        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/editPoolMember",
            method="POST",
            data=ET.tostring(request),
        ).object

        response_code = findtext(result, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def ex_destroy_pool_member(self, member, destroy_node=False):
        """
        Destroy a specific member of a pool

        :param pool: The instance of a pool member
        :type  pool: ``DimensionDataPoolMember``

        :param destroy_node: Also destroy the associated node
        :type  destroy_node: ``bool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        # remove the pool member
        destroy_request = ET.Element("removePoolMember", {"xmlns": TYPES_URN, "id": member.id})

        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/removePoolMember",
            method="POST",
            data=ET.tostring(destroy_request),
        ).object

        if member.node_id is not None and destroy_node is True:
            return self.ex_destroy_node(member.node_id)
        else:
            response_code = findtext(result, "responseCode", TYPES_URN)
            return response_code in ["IN_PROGRESS", "OK"]

    def ex_get_nodes(self, ex_network_domain_id=None):
        """
        Get the nodes within this geography or in given network.

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :return: Returns an ``list`` of ``DimensionDataVIPNode``
        :rtype: ``list`` of ``DimensionDataVIPNode``
        """
        params = None
        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        nodes = self.connection.request_with_orgId_api_2(
            "networkDomainVip/node", params=params
        ).object
        return self._to_nodes(nodes)

    def ex_get_node(self, node_id):
        """
        Get the node specified by node_id

        :return: Returns an instance of ``DimensionDataVIPNode``
        :rtype: Instance of ``DimensionDataVIPNode``
        """
        nodes = self.connection.request_with_orgId_api_2(
            "networkDomainVip/node/%s" % node_id
        ).object
        return self._to_node(nodes)

    def ex_destroy_node(self, node_id):
        """
        Destroy a specific node

        :param node_id: The ID of of a ``DimensionDataVIPNode``
        :type  node_id: ``str``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        # Destroy the node
        destroy_request = ET.Element("deleteNode", {"xmlns": TYPES_URN, "id": node_id})

        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/deleteNode",
            method="POST",
            data=ET.tostring(destroy_request),
        ).object
        response_code = findtext(result, "responseCode", TYPES_URN)
        return response_code in ["IN_PROGRESS", "OK"]

    def ex_wait_for_state(self, state, func, poll_interval=2, timeout=60, *args, **kwargs):
        """
        Wait for the function which returns a instance
        with field status to match

        Keep polling func until one of the desired states is matched

        :param state: Either the desired state (`str`) or a `list` of states
        :type  state: ``str`` or ``list``

        :param  func: The function to call, e.g. ex_get_vlan
        :type   func: ``function``

        :param  poll_interval: The number of seconds to wait between checks
        :type   poll_interval: `int`

        :param  timeout: The total number of seconds to wait to reach a state
        :type   timeout: `int`

        :param  args: The arguments for func
        :type   args: Positional arguments

        :param  kwargs: The arguments for func
        :type   kwargs: Keyword arguments
        """
        return self.connection.wait_for_state(state, func, poll_interval, timeout, *args, **kwargs)

    def ex_get_default_health_monitors(self, network_domain_id):
        """
        Get the default health monitors available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataDefaultHealthMonitor`
        """
        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/defaultHealthMonitor",
            params={"networkDomainId": network_domain_id},
            method="GET",
        ).object
        return self._to_health_monitors(result)

    def ex_get_default_persistence_profiles(self, network_domain_id):
        """
        Get the default persistence profiles available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataPersistenceProfile`
        """
        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/defaultPersistenceProfile",
            params={"networkDomainId": network_domain_id},
            method="GET",
        ).object
        return self._to_persistence_profiles(result)

    def ex_get_default_irules(self, network_domain_id):
        """
        Get the default iRules available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataDefaultiRule`
        """
        result = self.connection.request_with_orgId_api_2(
            action="networkDomainVip/defaultIrule",
            params={"networkDomainId": network_domain_id},
            method="GET",
        ).object
        return self._to_irules(result)

    def _to_irules(self, object):
        irules = []
        matches = object.findall(fixxpath("defaultIrule", TYPES_URN))
        for element in matches:
            irules.append(self._to_irule(element))
        return irules

    def _to_irule(self, element):
        compatible = []
        matches = element.findall(fixxpath("virtualListenerCompatibility", TYPES_URN))
        for match_element in matches:
            compatible.append(
                DimensionDataVirtualListenerCompatibility(
                    type=match_element.get("type"),
                    protocol=match_element.get("protocol", None),
                )
            )
        irule_element = element.find(fixxpath("irule", TYPES_URN))
        return DimensionDataDefaultiRule(
            id=irule_element.get("id"),
            name=irule_element.get("name"),
            compatible_listeners=compatible,
        )

    def _to_persistence_profiles(self, object):
        profiles = []
        matches = object.findall(fixxpath("defaultPersistenceProfile", TYPES_URN))
        for element in matches:
            profiles.append(self._to_persistence_profile(element))
        return profiles

    def _to_persistence_profile(self, element):
        compatible = []
        matches = element.findall(fixxpath("virtualListenerCompatibility", TYPES_URN))
        for match_element in matches:
            compatible.append(
                DimensionDataVirtualListenerCompatibility(
                    type=match_element.get("type"),
                    protocol=match_element.get("protocol", None),
                )
            )

        return DimensionDataPersistenceProfile(
            id=element.get("id"),
            fallback_compatible=bool(element.get("fallbackCompatible") == "true"),
            name=findtext(element, "name", TYPES_URN),
            compatible_listeners=compatible,
        )

    def _to_health_monitors(self, object):
        monitors = []
        matches = object.findall(fixxpath("defaultHealthMonitor", TYPES_URN))
        for element in matches:
            monitors.append(self._to_health_monitor(element))
        return monitors

    def _to_health_monitor(self, element):
        return DimensionDataDefaultHealthMonitor(
            id=element.get("id"),
            name=findtext(element, "name", TYPES_URN),
            node_compatible=bool(findtext(element, "nodeCompatible", TYPES_URN) == "true"),
            pool_compatible=bool(findtext(element, "poolCompatible", TYPES_URN) == "true"),
        )

    def _to_nodes(self, object):
        nodes = []
        for element in object.findall(fixxpath("node", TYPES_URN)):
            nodes.append(self._to_node(element))

        return nodes

    def _to_node(self, element):
        ipaddress = findtext(element, "ipv4Address", TYPES_URN)
        if ipaddress is None:
            ipaddress = findtext(element, "ipv6Address", TYPES_URN)

        name = findtext(element, "name", TYPES_URN)

        node = DimensionDataVIPNode(
            id=element.get("id"),
            name=name,
            status=self._VALUE_TO_STATE_MAP.get(
                findtext(element, "state", TYPES_URN), State.UNKNOWN
            ),
            connection_rate_limit=findtext(element, "connectionRateLimit", TYPES_URN),
            connection_limit=findtext(element, "connectionLimit", TYPES_URN),
            ip=ipaddress,
        )

        return node

    def _to_balancers(self, object):
        loadbalancers = []
        for element in object.findall(fixxpath("virtualListener", TYPES_URN)):
            loadbalancers.append(self._to_balancer(element))

        return loadbalancers

    def _to_balancer(self, element):
        ipaddress = findtext(element, "listenerIpAddress", TYPES_URN)
        name = findtext(element, "name", TYPES_URN)
        port = findtext(element, "port", TYPES_URN)
        extra = {}

        pool_element = element.find(fixxpath("pool", TYPES_URN))
        if pool_element is None:
            extra["pool_id"] = None

        else:
            extra["pool_id"] = pool_element.get("id")

        extra["network_domain_id"] = findtext(element, "networkDomainId", TYPES_URN)

        balancer = LoadBalancer(
            id=element.get("id"),
            name=name,
            state=self._VALUE_TO_STATE_MAP.get(
                findtext(element, "state", TYPES_URN), State.UNKNOWN
            ),
            ip=ipaddress,
            port=port,
            driver=self.connection.driver,
            extra=extra,
        )

        return balancer

    def _to_members(self, object):
        members = []
        for element in object.findall(fixxpath("poolMember", TYPES_URN)):
            members.append(self._to_member(element))

        return members

    def _to_member(self, element):
        port = findtext(element, "port", TYPES_URN)
        if port is not None:
            port = int(port)
        pool_member = DimensionDataPoolMember(
            id=element.get("id"),
            name=element.find(fixxpath("node", TYPES_URN)).get("name"),
            status=findtext(element, "state", TYPES_URN),
            node_id=element.find(fixxpath("node", TYPES_URN)).get("id"),
            ip=element.find(fixxpath("node", TYPES_URN)).get("ipAddress"),
            port=port,
        )
        return pool_member

    def _to_pools(self, object):
        pools = []
        for element in object.findall(fixxpath("pool", TYPES_URN)):
            pools.append(self._to_pool(element))

        return pools

    def _to_pool(self, element):
        pool = DimensionDataPool(
            id=element.get("id"),
            name=findtext(element, "name", TYPES_URN),
            status=findtext(element, "state", TYPES_URN),
            description=findtext(element, "description", TYPES_URN),
            load_balance_method=findtext(element, "loadBalanceMethod", TYPES_URN),
            health_monitor_id=findtext(element, "healthMonitorId", TYPES_URN),
            service_down_action=findtext(element, "serviceDownAction", TYPES_URN),
            slow_ramp_time=findtext(element, "slowRampTime", TYPES_URN),
        )
        return pool



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import zlib
from io import StringIO
from unittest import mock

import requests_mock

import libcloud
from libcloud.http import LibcloudConnection
from libcloud.test import unittest
from libcloud.common.base import Connection
from libcloud.utils.loggingconnection import LoggingConnection

EXPECTED_DATA_JSON = """
HTTP/1.1 200 OK
Content-Type: application/json

{"foo": "bar!"}
""".strip()

EXPECTED_DATA_JSON_PRETTY = """
HTTP/1.1 200 OK
Content-Type: application/json

{
    "foo": "bar!"
}
""".strip()

EXPECTED_DATA_XML = """
HTTP/1.1 200 OK
Content-Type: text/xml

<foo><bar /></foo>
""".strip()

EXPECTED_DATA_XML_PRETTY = """
HTTP/1.1 200 OK
Content-Type: application/xml

<foo><bar /></foo>
""".strip()


class TestLoggingConnection(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self._reset_environ()

    def tearDown(self):
        super().tearDown()
        Connection.conn_class = LibcloudConnection

    def test_debug_method_uses_log_class(self):
        with StringIO() as fh:
            libcloud.enable_debug(fh)
            conn = Connection(timeout=10)
            conn.connect()
        self.assertTrue(isinstance(conn.connection, LoggingConnection))

    def test_debug_log_class_handles_request(self):
        with StringIO() as fh:
            libcloud.enable_debug(fh)
            conn = Connection(url="http://test.com/")
            conn.connect()
            self.assertEqual(conn.connection.host, "http://test.com")
            with requests_mock.mock() as m:
                m.get("http://test.com/test", text="data")
                conn.request("/test")
            log = fh.getvalue()
        self.assertTrue(isinstance(conn.connection, LoggingConnection))
        self.assertIn("-i -X GET", log)
        self.assertIn("data", log)

    def test_debug_log_class_handles_request_with_compression(self):
        request = zlib.compress(b"data")
        with StringIO() as fh:
            libcloud.enable_debug(fh)
            conn = Connection(url="http://test.com/")
            conn.connect()
            self.assertEqual(conn.connection.host, "http://test.com")
            with requests_mock.mock() as m:
                m.get(
                    "http://test.com/test",
                    content=request,
                    headers={"content-encoding": "zlib"},
                )
                conn.request("/test")
            log = fh.getvalue()
        self.assertTrue(isinstance(conn.connection, LoggingConnection))
        self.assertIn("-i -X GET", log)

    def test_log_response_json_content_type(self):
        conn = LoggingConnection(host="example.com", port=80)

        r = self._get_mock_response("application/json", '{"foo": "bar!"}')
        result = conn._log_response(r).replace("\r", "")
        self.assertTrue(EXPECTED_DATA_JSON in result)

    def test_log_response_xml_content_type(self):
        conn = LoggingConnection(host="example.com", port=80)

        r = self._get_mock_response("text/xml", "<foo><bar /></foo>")
        result = conn._log_response(r).replace("\r", "")
        self.assertTrue(EXPECTED_DATA_XML in result)

    def test_log_response_with_pretty_print_json_content_type(self):
        os.environ["LIBCLOUD_DEBUG_PRETTY_PRINT_RESPONSE"] = "1"

        conn = LoggingConnection(host="example.com", port=80)

        # body type is unicode
        r = self._get_mock_response("application/json", '{"foo": "bar!"}')
        result = conn._log_response(r).replace("\r", "")
        self.assertTrue(EXPECTED_DATA_JSON_PRETTY in result)

        # body type is bytes
        r = self._get_mock_response("application/json", bytes('{"foo": "bar!"}', "utf-8"))
        result = conn._log_response(r).replace("\r", "")
        self.assertTrue(EXPECTED_DATA_JSON_PRETTY in result)

    def test_log_response_with_pretty_print_xml_content_type(self):
        os.environ["LIBCLOUD_DEBUG_PRETTY_PRINT_RESPONSE"] = "1"

        conn = LoggingConnection(host="example.com", port=80)

        r = self._get_mock_response("application/xml", "<foo><bar /></foo>")
        result = conn._log_response(r).replace("\r", "")
        self.assertTrue(EXPECTED_DATA_XML_PRETTY in result)

    def _reset_environ(self):
        if "LIBCLOUD_DEBUG_PRETTY_PRINT_RESPONSE" in os.environ:
            del os.environ["LIBCLOUD_DEBUG_PRETTY_PRINT_RESPONSE"]

    def _get_mock_response(self, content_type, body):
        header = mock.Mock()
        header.title.return_value = "Content-Type"
        header.lower.return_value = "content-type"

        r = mock.Mock()
        r.version = 11
        r.status = "200"
        r.reason = "OK"
        r.getheaders.return_value = [(header, content_type)]
        r.read.return_value = body

        return r


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import errno
import hashlib
from io import BytesIO
from unittest import mock
from unittest.mock import Mock

from libcloud.test import MockHttp, BodyStream, unittest
from libcloud.utils.py3 import PY2, StringIO, b, httplib, assertRaisesRegex
from libcloud.storage.base import DEFAULT_CONTENT_TYPE, StorageDriver
from libcloud.common.exceptions import RateLimitReachedError
from libcloud.test.storage.base import BaseRangeDownloadMockHttp


class BaseMockRawResponse(MockHttp):
    def _(self, method, url, body, headers):
        body = "ab"
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def root(self, method, url, body, headers):
        body = "ab"
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


class BaseStorageTests(unittest.TestCase):
    def setUp(self):
        self.send_called = 0
        StorageDriver.connectionCls.conn_class = BaseMockRawResponse

        self.driver1 = StorageDriver("username", "key", host="localhost")
        self.driver1.supports_chunked_encoding = True

        self.driver2 = StorageDriver("username", "key", host="localhost")
        self.driver2.supports_chunked_encoding = False

        self.driver1.strict_mode = False
        self.driver1.strict_mode = False

    def test__upload_object_iterator_must_have_next_method(self):
        valid_iterators = [BytesIO(b("134")), StringIO("bar")]
        invalid_iterators = ["foobar", "", False, True, 1, object()]

        kwargs = {
            "object_name": "foo",
            "content_type": "foo/bar",
            "request_path": "/",
            "headers": {},
        }

        for value in valid_iterators:
            kwargs["stream"] = value
            self.driver1._upload_object(**kwargs)

        for value in invalid_iterators:
            kwargs["stream"] = value

            try:
                self.driver1._upload_object(**kwargs)
            except AttributeError:
                pass
            else:
                self.fail("Exception was not thrown")

    def test__upload_object_does_not_stream_response(self):
        resp = self.driver1._upload_object(
            object_name="foo",
            content_type="foo/bar",
            request_path="/",
            stream=iter(b"foo"),
        )
        mock_response = resp["response"].response._response
        response_streamed = mock_response.request.stream
        assert response_streamed is False

    def test__get_hash_function(self):
        self.driver1.hash_type = "md5"
        func = self.driver1._get_hash_function()
        self.assertTrue(func)

        self.driver1.hash_type = "sha1"
        func = self.driver1._get_hash_function()
        self.assertTrue(func)

        try:
            self.driver1.hash_type = "invalid-hash-function"
            func = self.driver1._get_hash_function()
        except RuntimeError:
            pass
        else:
            self.fail("Invalid hash type but exception was not thrown")

    def test_upload_no_content_type_supplied_or_detected(self):
        iterator = StringIO()

        # strict_mode is disabled, default content type should be used
        self.driver1.connection = Mock()

        self.driver1._upload_object(
            object_name="test", content_type=None, request_path="/", stream=iterator
        )

        headers = self.driver1.connection.request.call_args[-1]["headers"]
        self.assertEqual(headers["Content-Type"], DEFAULT_CONTENT_TYPE)

        # strict_mode is enabled, exception should be thrown

        self.driver1.strict_mode = True
        expected_msg = (
            'File content-type could not be guessed for "test" '
            "and no content_type value is provided"
        )
        assertRaisesRegex(
            self,
            AttributeError,
            expected_msg,
            self.driver1._upload_object,
            object_name="test",
            content_type=None,
            request_path="/",
            stream=iterator,
        )

    @mock.patch("libcloud.utils.files.exhaust_iterator")
    @mock.patch("libcloud.utils.files.read_in_chunks")
    def test_upload_object_hash_calculation_is_efficient(
        self, mock_read_in_chunks, mock_exhaust_iterator
    ):
        # Verify that we don't buffer whole file in memory when calculating
        # object has when iterator has __next__ method, but instead read and
        # calculate hash in chunks
        size = 100

        self.driver1.connection = Mock()

        # stream has __next__ method and next() method
        mock_read_in_chunks.return_value = "a" * size

        iterator = BodyStream("a" * size)
        self.assertTrue(hasattr(iterator, "__next__"))
        self.assertTrue(hasattr(iterator, "next"))

        self.assertEqual(mock_read_in_chunks.call_count, 0)
        self.assertEqual(mock_exhaust_iterator.call_count, 0)

        result = self.driver1._upload_object(
            object_name="test1", content_type=None, request_path="/", stream=iterator
        )

        hasher = hashlib.md5()
        hasher.update(b("a") * size)
        expected_hash = hasher.hexdigest()

        self.assertEqual(result["data_hash"], expected_hash)
        self.assertEqual(result["bytes_transferred"], size)

        headers = self.driver1.connection.request.call_args[-1]["headers"]
        self.assertEqual(headers["Content-Type"], DEFAULT_CONTENT_TYPE)

        self.assertEqual(mock_read_in_chunks.call_count, 1)
        self.assertEqual(mock_exhaust_iterator.call_count, 0)

        # stream has only has next() method
        mock_read_in_chunks.return_value = "b" * size

        iterator = iter([str(v) for v in ["b" * size]])

        if PY2:
            self.assertFalse(hasattr(iterator, "__next__"))
            self.assertTrue(hasattr(iterator, "next"))
        else:
            self.assertTrue(hasattr(iterator, "__next__"))
            self.assertFalse(hasattr(iterator, "next"))

        self.assertEqual(mock_read_in_chunks.call_count, 1)
        self.assertEqual(mock_exhaust_iterator.call_count, 0)

        self.assertEqual(mock_read_in_chunks.call_count, 1)
        self.assertEqual(mock_exhaust_iterator.call_count, 0)

        result = self.driver1._upload_object(
            object_name="test2", content_type=None, request_path="/", stream=iterator
        )

        hasher = hashlib.md5()
        hasher.update(b("b") * size)
        expected_hash = hasher.hexdigest()

        self.assertEqual(result["data_hash"], expected_hash)
        self.assertEqual(result["bytes_transferred"], size)

        headers = self.driver1.connection.request.call_args[-1]["headers"]
        self.assertEqual(headers["Content-Type"], DEFAULT_CONTENT_TYPE)

        self.assertEqual(mock_read_in_chunks.call_count, 2)
        self.assertEqual(mock_exhaust_iterator.call_count, 0)

    def test_upload_object_via_stream_illegal_seek_errors_are_ignored(self):
        # Illegal seek errors should be ignored
        size = 100

        self.driver1.connection = Mock()

        seek_error = OSError("Illegal seek")
        seek_error.errno = 29
        assert errno.ESPIPE == 29

        iterator = BodyStream("a" * size)
        iterator.seek = mock.Mock(side_effect=seek_error)

        result = self.driver1._upload_object(
            object_name="test1", content_type=None, request_path="/", stream=iterator
        )

        hasher = hashlib.md5()
        hasher.update(b("a") * size)
        expected_hash = hasher.hexdigest()

        self.assertEqual(result["data_hash"], expected_hash)
        self.assertEqual(result["bytes_transferred"], size)

        # But others shouldn't
        self.driver1.connection = Mock()

        seek_error = OSError("Other error")
        seek_error.errno = 21

        iterator = BodyStream("b" * size)
        iterator.seek = mock.Mock(side_effect=seek_error)

        self.assertRaisesRegex(
            OSError,
            "Other error",
            self.driver1._upload_object,
            object_name="test1",
            content_type=None,
            request_path="/",
            stream=iterator,
        )

    def test_get_standard_range_str(self):
        result = self.driver1._get_standard_range_str(0, 5)
        self.assertEqual(result, "bytes=0-4")

        result = self.driver1._get_standard_range_str(0)
        self.assertEqual(result, "bytes=0-")
        result = self.driver1._get_standard_range_str(0, 1)

        self.assertEqual(result, "bytes=0-0")

        result = self.driver1._get_standard_range_str(200)
        self.assertEqual(result, "bytes=200-")

        result = self.driver1._get_standard_range_str(10, 200)
        self.assertEqual(result, "bytes=10-199")

        result = self.driver1._get_standard_range_str(10, 11)
        self.assertEqual(result, "bytes=10-10")

        result = self.driver1._get_standard_range_str(10, 11, True)
        self.assertEqual(result, "bytes=10-11")

    @mock.patch("os.environ", {"LIBCLOUD_RETRY_FAILED_HTTP_REQUESTS": True})
    def test_should_retry_rate_limited_errors(self):
        class SecondException(Exception):
            pass

        count = 0

        def raise_on_second(*_, **__):
            nonlocal count
            count += 1
            if count > 1:
                raise SecondException()
            else:
                raise RateLimitReachedError()

        send_mock = Mock()
        self.driver1.connection.connection.session.send = send_mock
        send_mock.side_effect = raise_on_second
        with self.assertRaises(SecondException):
            self.driver1._upload_object(
                object_name="some name",
                content_type="something",
                request_path="/",
                stream=iter([]),
            )

    @mock.patch("os.environ", {"LIBCLOUD_RETRY_FAILED_HTTP_REQUESTS": True})
    def test_should_retry_rate_limited_errors_until_success(self):
        count = 0

        def succeed_on_second(*_, **__) -> mock.MagicMock:
            nonlocal count
            count += 1
            if count > 1:
                successful_response = mock.MagicMock()
                successful_response.status_code = 200
                return successful_response
            else:
                raise RateLimitReachedError()

        self.driver1.connection.connection.session.send = Mock(side_effect=succeed_on_second)
        uploaded_object = self.driver1._upload_object(
            object_name="some name",
            content_type="something",
            request_path="/",
            stream=iter([]),
        )

        self.assertEqual(
            True,
            uploaded_object["response"].success(),
            "Expected to have successful response after retry",
        )


class BaseRangeDownloadMockHttpTestCase(unittest.TestCase):
    def test_get_start_and_end_bytes_from_range_str(self):
        mock_http = BaseRangeDownloadMockHttp(None, None)
        body = "0123456789"

        range_str = "bytes=1-"
        result = mock_http._get_start_and_end_bytes_from_range_str(range_str, body)
        self.assertEqual(result, (1, len(body)))
        range_str = "bytes=1-5"

        result = mock_http._get_start_and_end_bytes_from_range_str(range_str, body)
        self.assertEqual(result, (1, 5))

        range_str = "bytes=3-5"
        result = mock_http._get_start_and_end_bytes_from_range_str(range_str, body)
        self.assertEqual(result, (3, 5))


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib, assertRaisesRegex
from libcloud.common.types import InvalidCredsError
from libcloud.compute.types import NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.cloudsigma import (
    CloudSigmaError,
    CloudSigmaNodeDriver,
    CloudSigma_2_0_NodeDriver,
)

try:
    import simplejson as json
except Exception:
    import json


class CloudSigmaAPI20BaseTestCase:
    def setUp(self):
        self.driver_klass.connectionCls.conn_class = CloudSigmaMockHttp

        CloudSigmaMockHttp.type = None
        CloudSigmaMockHttp.use_param = "do"

        self.driver = self.driver_klass(*self.driver_args, **self.driver_kwargs)
        self.driver.DRIVE_TRANSITION_SLEEP_INTERVAL = 0.1
        self.driver.DRIVE_TRANSITION_TIMEOUT = 1
        self.node = self.driver.list_nodes()[0]

    def test_invalid_api_versions(self):
        expected_msg = "Unsupported API version: invalid"
        assertRaisesRegex(
            self,
            NotImplementedError,
            expected_msg,
            CloudSigmaNodeDriver,
            "username",
            "password",
            api_version="invalid",
        )

    def test_invalid_credentials(self):
        CloudSigmaMockHttp.type = "INVALID_CREDS"
        self.assertRaises(InvalidCredsError, self.driver.list_nodes)

    def test_invalid_region(self):
        expected_msg = "Invalid region:"
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            CloudSigma_2_0_NodeDriver,
            "foo",
            "bar",
            region="invalid",
        )

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()

        size = sizes[0]
        self.assertEqual(size.id, "small-1")

    def test_list_images(self):
        images = self.driver.list_images()

        image = images[0]
        self.assertEqual(image.name, "ubuntu-10.04-toMP")
        self.assertEqual(image.extra["image_type"], "preinst")
        self.assertEqual(image.extra["media"], "disk")
        self.assertEqual(image.extra["os"], "linux")

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()

        node = nodes[0]
        self.assertEqual(len(nodes), 2)
        self.assertEqual(node.id, "9de75ed6-fd33-45e2-963f-d405f31fd911")
        self.assertEqual(node.name, "test no drives")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.public_ips, ["185.12.5.181", "178.22.68.55"])
        self.assertEqual(node.private_ips, [])

    def test_create_node(self):
        image = self.driver.list_images()[0]
        size = self.driver.list_sizes()[0]
        metadata = {"foo": "bar"}

        node = self.driver.create_node(
            name="test node", size=size, image=image, ex_metadata=metadata
        )
        self.assertEqual(node.name, "test node")
        self.assertEqual(len(node.extra["nics"]), 1)
        self.assertEqual(node.extra["nics"][0]["ip_v4_conf"]["conf"], "dhcp")

    def test_create_node_with_vlan(self):
        image = self.driver.list_images()[0]
        size = self.driver.list_sizes()[0]

        vlan_uuid = "39ae851d-433f-4ac2-a803-ffa24cb1fa3e"

        node = self.driver.create_node(
            name="test node vlan", size=size, image=image, ex_vlan=vlan_uuid
        )
        self.assertEqual(node.name, "test node vlan")
        self.assertEqual(len(node.extra["nics"]), 2)
        self.assertEqual(node.extra["nics"][0]["ip_v4_conf"]["conf"], "dhcp")
        self.assertEqual(node.extra["nics"][1]["vlan"]["uuid"], vlan_uuid)

    def test_destroy_node(self):
        status = self.driver.destroy_node(node=self.node)
        self.assertTrue(status)

    def test_ex_start_node(self):
        status = self.driver.ex_start_node(node=self.node)
        self.assertTrue(status)

    def test_ex_start_node_avoid_mode(self):
        CloudSigmaMockHttp.type = "AVOID_MODE"
        ex_avoid = ["1", "2"]
        status = self.driver.ex_start_node(node=self.node, ex_avoid=ex_avoid)
        self.assertTrue(status)

    def test_ex_start_node_already_started(self):
        CloudSigmaMockHttp.type = "ALREADY_STARTED"

        expected_msg = 'Cannot start guest in state "started". Guest should ' 'be in state "stopped'

        assertRaisesRegex(
            self,
            CloudSigmaError,
            expected_msg,
            self.driver.ex_start_node,
            node=self.node,
        )

    def test_ex_stop_node(self):
        status = self.driver.ex_stop_node(node=self.node)
        self.assertTrue(status)

    def test_ex_stop_node_already_stopped(self):
        CloudSigmaMockHttp.type = "ALREADY_STOPPED"

        expected_msg = 'Cannot stop guest in state "stopped"'
        assertRaisesRegex(
            self,
            CloudSigmaError,
            expected_msg,
            self.driver.ex_stop_node,
            node=self.node,
        )

    def test_reboot_node(self):
        status = self.driver.reboot_node(node=self.node)
        self.assertTrue(status)

    def test_ex_get_node(self):
        node_id = "3df825cb-9c1b-470d-acbd-03e1a966c046"
        node = self.driver.ex_get_node(node_id)

        self.assertEqual(node.id, "3df825cb-9c1b-470d-acbd-03e1a966c046")
        self.assertEqual(node.name, "test_server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.public_ips, ["162.213.38.202"])
        self.assertEqual(node.private_ips, [])

    def test_ex_clone_node(self):
        node_to_clone = self.driver.list_nodes()[0]

        cloned_node = self.driver.ex_clone_node(node=node_to_clone, name="test cloned node")
        self.assertEqual(cloned_node.name, "test cloned node")

    def test_ex_open_vnc_tunnel(self):
        node = self.driver.list_nodes()[0]
        vnc_url = self.driver.ex_open_vnc_tunnel(node=node)
        self.assertEqual(vnc_url, "vnc://direct.lvs.cloudsigma.com:41111")

    def test_ex_close_vnc_tunnel(self):
        node = self.driver.list_nodes()[0]
        status = self.driver.ex_close_vnc_tunnel(node=node)
        self.assertTrue(status)

    def test_ex_list_library_drives(self):
        drives = self.driver.ex_list_library_drives()

        drive = drives[0]
        self.assertEqual(drive.name, "IPCop 2.0.2")
        self.assertEqual(drive.size, 1)
        self.assertEqual(drive.media, "cdrom")
        self.assertEqual(drive.status, "unmounted")

    def test_ex_list_user_drives(self):
        drives = self.driver.ex_list_user_drives()

        drive = drives[0]
        self.assertEqual(drive.name, "test node 2-drive")
        self.assertEqual(drive.size, 13)
        self.assertEqual(drive.media, "disk")
        self.assertEqual(drive.status, "unmounted")

    def test_ex_create_drive(self):
        CloudSigmaMockHttp.type = "CREATE"

        name = "test drive 5"
        size = 2000 * 1024 * 1024

        drive = self.driver.ex_create_drive(name=name, size=size, media="disk")
        self.assertEqual(drive.name, "test drive 5")
        self.assertEqual(drive.media, "disk")

    def test_ex_clone_drive(self):
        drive = self.driver.ex_list_user_drives()[0]
        cloned_drive = self.driver.ex_clone_drive(drive=drive, name="cloned drive")

        self.assertEqual(cloned_drive.name, "cloned drive")

    def test_ex_resize_drive(self):
        drive = self.driver.ex_list_user_drives()[0]

        size = 10

        resized_drive = self.driver.ex_resize_drive(drive=drive, size=size)
        self.assertEqual(resized_drive.name, "test drive 5")
        self.assertEqual(resized_drive.media, "disk")
        self.assertEqual(resized_drive.size, size)

    def test_ex_list_firewall_policies(self):
        policies = self.driver.ex_list_firewall_policies()

        policy = policies[1]
        rule = policy.rules[0]
        self.assertEqual(policy.name, "My awesome policy")
        self.assertEqual(rule.action, "drop")
        self.assertEqual(rule.direction, "out")
        self.assertEqual(rule.dst_ip, "23.0.0.0/32")
        self.assertEqual(rule.ip_proto, "tcp")
        self.assertIsNone(rule.dst_port)
        self.assertIsNone(rule.src_ip)
        self.assertIsNone(rule.src_port)
        self.assertEqual(rule.comment, "Drop traffic from the VM to IP address 23.0.0.0/32")

    def test_ex_create_firewall_policy_no_rules(self):
        CloudSigmaMockHttp.type = "CREATE_NO_RULES"
        policy = self.driver.ex_create_firewall_policy(name="test policy 1")

        self.assertEqual(policy.name, "test policy 1")
        self.assertEqual(policy.rules, [])

    def test_ex_create_firewall_policy_with_rules(self):
        CloudSigmaMockHttp.type = "CREATE_WITH_RULES"
        rules = [
            {
                "action": "accept",
                "direction": "out",
                "ip_proto": "tcp",
                "src_ip": "127.0.0.1",
                "dst_ip": "127.0.0.1",
            }
        ]

        policy = self.driver.ex_create_firewall_policy(name="test policy 2", rules=rules)
        rule = policy.rules[0]

        self.assertEqual(policy.name, "test policy 2")
        self.assertEqual(len(policy.rules), 1)

        self.assertEqual(rule.action, "accept")
        self.assertEqual(rule.direction, "out")
        self.assertEqual(rule.ip_proto, "tcp")

    def test_ex_attach_firewall_policy(self):
        policy = self.driver.ex_list_firewall_policies()[0]
        node = self.driver.list_nodes()[0]

        CloudSigmaMockHttp.type = "ATTACH_POLICY"
        updated_node = self.driver.ex_attach_firewall_policy(policy=policy, node=node)
        nic = updated_node.extra["nics"][0]
        self.assertEqual(nic["firewall_policy"]["uuid"], "461dfb8c-e641-43d7-a20e-32e2aa399086")

    def test_ex_attach_firewall_policy_inexistent_nic(self):
        policy = self.driver.ex_list_firewall_policies()[0]
        node = self.driver.list_nodes()[0]

        nic_mac = "inexistent"
        expected_msg = "Cannot find the NIC interface to attach a policy to"
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            self.driver.ex_attach_firewall_policy,
            policy=policy,
            node=node,
            nic_mac=nic_mac,
        )

    def test_ex_delete_firewall_policy(self):
        policy = self.driver.ex_list_firewall_policies()[0]
        status = self.driver.ex_delete_firewall_policy(policy=policy)
        self.assertTrue(status)

    def test_ex_list_tags(self):
        tags = self.driver.ex_list_tags()

        tag = tags[0]
        self.assertEqual(tag.id, "a010ec41-2ead-4630-a1d0-237fa77e4d4d")
        self.assertEqual(tag.name, "test tag 2")

    def test_ex_get_tag(self):
        tag = self.driver.ex_get_tag(tag_id="a010ec41-2ead-4630-a1d0-237fa77e4d4d")

        self.assertEqual(tag.id, "a010ec41-2ead-4630-a1d0-237fa77e4d4d")
        self.assertEqual(tag.name, "test tag 2")

    def test_ex_create_tag(self):
        tag = self.driver.ex_create_tag(name="test tag 3")
        self.assertEqual(tag.name, "test tag 3")

    def test_ex_create_tag_with_resources(self):
        CloudSigmaMockHttp.type = "WITH_RESOURCES"
        resource_uuids = ["1"]
        tag = self.driver.ex_create_tag(name="test tag 3", resource_uuids=resource_uuids)
        self.assertEqual(tag.name, "test tag 3")
        self.assertEqual(tag.resources, resource_uuids)

    def test_ex_tag_resource(self):
        node = self.driver.list_nodes()[0]
        tag = self.driver.ex_list_tags()[0]

        updated_tag = self.driver.ex_tag_resource(resource=node, tag=tag)
        self.assertEqual(updated_tag.name, "test tag 3")

    def test_ex_tag_resources(self):
        nodes = self.driver.list_nodes()
        tag = self.driver.ex_list_tags()[0]

        updated_tag = self.driver.ex_tag_resources(resources=nodes, tag=tag)
        self.assertEqual(updated_tag.name, "test tag 3")

    def test_ex_tag_resource_invalid_resource_object(self):
        tag = self.driver.ex_list_tags()[0]

        expected_msg = "Resource doesn't have id attribute"
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            self.driver.ex_tag_resource,
            tag=tag,
            resource={},
        )

    def test_ex_delete_tag(self):
        tag = self.driver.ex_list_tags()[0]
        status = self.driver.ex_delete_tag(tag=tag)
        self.assertTrue(status)

    def test_ex_get_balance(self):
        balance = self.driver.ex_get_balance()
        self.assertEqual(balance["balance"], "10.00")
        self.assertEqual(balance["currency"], "USD")

    def test_ex_get_pricing(self):
        pricing = self.driver.ex_get_pricing()

        self.assertTrue("current" in pricing)
        self.assertTrue("next" in pricing)
        self.assertTrue("objects" in pricing)

    def test_ex_get_usage(self):
        pricing = self.driver.ex_get_usage()

        self.assertTrue("balance" in pricing)
        self.assertTrue("usage" in pricing)

    def test_ex_list_subscriptions(self):
        subscriptions = self.driver.ex_list_subscriptions()

        self.assertEqual(len(subscriptions), 6)

        subscription = subscriptions[0]
        self.assertEqual(subscription.id, "7272")
        self.assertEqual(subscription.resource, "vlan")
        self.assertEqual(subscription.amount, 1)
        self.assertEqual(subscription.period, "345 days, 0:00:00")
        self.assertEqual(subscription.status, "active")
        self.assertEqual(subscription.price, "0E-20")

        subscription = subscriptions[-1]
        self.assertEqual(subscription.id, "5555")
        self.assertEqual(subscription.start_time, None)
        self.assertEqual(subscription.end_time, None)

    def test_ex_create_subscription(self):
        CloudSigmaMockHttp.type = "CREATE_SUBSCRIPTION"
        subscription = self.driver.ex_create_subscription(
            amount=1, period="1 month", resource="vlan"
        )
        self.assertEqual(subscription.amount, 1)
        self.assertEqual(subscription.period, "1 month")
        self.assertEqual(subscription.resource, "vlan")
        self.assertEqual(subscription.price, "10.26666666666666666666666667")
        self.assertEqual(subscription.auto_renew, False)
        self.assertEqual(subscription.subscribed_object, "2494079f-8376-40bf-9b37-34d633b8a7b7")

    def test_ex_list_subscriptions_status_filterting(self):
        CloudSigmaMockHttp.type = "STATUS_FILTER"
        self.driver.ex_list_subscriptions(status="active")

    def test_ex_list_subscriptions_resource_filterting(self):
        CloudSigmaMockHttp.type = "RESOURCE_FILTER"
        resources = ["cpu", "mem"]
        self.driver.ex_list_subscriptions(resources=resources)

    def test_ex_toggle_subscription_auto_renew(self):
        subscription = self.driver.ex_list_subscriptions()[0]
        status = self.driver.ex_toggle_subscription_auto_renew(subscription=subscription)
        self.assertTrue(status)

    def test_ex_list_capabilities(self):
        capabilities = self.driver.ex_list_capabilities()
        self.assertEqual(capabilities["servers"]["cpu"]["min"], 250)

    def test_ex_list_servers_availability_groups(self):
        groups = self.driver.ex_list_servers_availability_groups()
        self.assertEqual(len(groups), 3)
        self.assertEqual(len(groups[0]), 2)
        self.assertEqual(len(groups[2]), 1)

    def test_ex_list_drives_availability_groups(self):
        groups = self.driver.ex_list_drives_availability_groups()
        self.assertEqual(len(groups), 1)
        self.assertEqual(len(groups[0]), 11)

    def test_ex_attach_drive(self):
        node_id = "3df825cb-9c1b-470d-acbd-03e1a966c046"
        node = self.driver.ex_get_node(node_id)
        drives = self.driver.ex_list_user_drives()
        drive = drives[0]
        response = self.driver.ex_attach_drive(node, drive)
        self.assertTrue(response)

    def test_ex_detach_drive(self):
        node_id = "3df825cb-9c1b-470d-acbd-03e1a966c046"
        node = self.driver.ex_get_node(node_id)
        drives = self.driver.ex_list_user_drives()
        drive = drives[0]
        response = self.driver.ex_detach_drive(node, drive)
        self.assertTrue(response)

    def test_wait_for_drive_state_transition_timeout(self):
        drive = self.driver.ex_list_user_drives()[0]
        state = "timeout"

        expected_msg = "Timed out while waiting for drive transition"
        assertRaisesRegex(
            self,
            Exception,
            expected_msg,
            self.driver._wait_for_drive_state_transition,
            drive=drive,
            state=state,
            timeout=0.5,
        )

    def test_wait_for_drive_state_transition_success(self):
        drive = self.driver.ex_list_user_drives()[0]
        state = "unmounted"

        drive = self.driver._wait_for_drive_state_transition(drive=drive, state=state, timeout=0.5)
        self.assertEqual(drive.status, state)

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()

        key = keys[0]
        self.assertEqual(len(keys), 2)
        self.assertEqual(key.name, "ala")
        self.assertEqual(key.fingerprint, "8c:71:a2:f0:bc:a4:45:66:a8:59:77:6d:80:d5:a0:89")
        self.assertEqual(key.extra["uuid"], "186106ac-afb5-40e5-a0de-6f0feba5a3d5")

    def test_import_key_pair_from_string(self):
        public_key = self.driver.list_key_pairs()[0].public_key
        key = self.driver.import_key_pair_from_string("test_key", public_key)
        self.assertEqual(key.name, "test_key")
        self.assertIsNone(key.private_key)
        self.assertEqual(key.extra["uuid"], "6d7db550-d1c5-4e3b-9c0a-e83f6a2b8244")
        self.assertEqual(key.public_key, public_key)

    def test_get_key_pair(self):
        uuid = self.driver.list_key_pairs()[0].extra["uuid"]
        key = self.driver.get_key_pair(uuid)
        self.assertEqual(key.name, "ala")
        self.assertEqual(key.extra["uuid"], "186106ac-afb5-40e5-a0de-6f0feba5a3d5")

    def test_delete_key_pair(self):
        key = self.driver.list_key_pairs()[0]
        response = self.driver.delete_key_pair(key)
        self.assertTrue(response)


class CloudSigmaAPI20DirectTestCase(CloudSigmaAPI20BaseTestCase, unittest.TestCase):
    driver_klass = CloudSigma_2_0_NodeDriver
    driver_args = ("foo", "bar")
    driver_kwargs = {}


class CloudSigmaAPI20IndirectTestCase(CloudSigmaAPI20BaseTestCase, unittest.TestCase):
    driver_klass = CloudSigmaNodeDriver
    driver_args = ("foo", "bar")
    driver_kwargs = {"api_version": "2.0"}


class CloudSigmaMockHttp(MockHttp, unittest.TestCase):
    fixtures = ComputeFileFixtures("cloudsigma_2_0")

    def _api_2_0_servers_detail_INVALID_CREDS(self, method, url, body, headers):
        body = self.fixtures.load("libdrives.json")
        return (httplib.UNAUTHORIZED, body, {}, httplib.responses[httplib.UNAUTHORIZED])

    def _api_2_0_libdrives(self, method, url, body, headers):
        body = self.fixtures.load("libdrives.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_servers_detail(self, method, url, body, headers):
        body = self.fixtures.load("servers_detail_mixed_state.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911(self, method, url, body, headers):
        body = ""
        return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _api_2_0_servers(self, method, url, body, headers):
        if method == "POST":
            # create_node

            parsed = json.loads(body)

            if "vlan" in parsed["name"]:
                self.assertEqual(len(parsed["nics"]), 2)
                body = self.fixtures.load("servers_create_with_vlan.json")
            else:
                body = self.fixtures.load("servers_create.json")

            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_start(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("start_success.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_AVOID_MODE_start(
        self, method, url, body, headers
    ):
        self.assertUrlContainsQueryParams(url, {"avoid": "1,2"})

        body = self.fixtures.load("start_success.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_ALREADY_STARTED_start(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("start_already_started.json")
        return (httplib.FORBIDDEN, body, {}, httplib.responses[httplib.FORBIDDEN])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_stop(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("stop_success.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_ALREADY_STOPPED_stop(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("stop_already_stopped.json")
        return (httplib.FORBIDDEN, body, {}, httplib.responses[httplib.FORBIDDEN])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_clone(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("servers_clone.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_open_vnc(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("servers_open_vnc.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_action_close_vnc(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("servers_close_vnc.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_servers_3df825cb_9c1b_470d_acbd_03e1a966c046(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("servers_get.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "PUT":
            body = ""
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_drives_detail(self, method, url, body, headers):
        body = self.fixtures.load("drives_detail.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_drives_b02311e2_a83c_4c12_af10_b30d51c86913(self, method, url, body, headers):
        body = self.fixtures.load("drives_get.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_drives_9d1d2cf3_08c1_462f_8485_f4b073560809(self, method, url, body, headers):
        body = self.fixtures.load("drives_get.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_drives_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("drives_create.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_drives_9d1d2cf3_08c1_462f_8485_f4b073560809_action_clone(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("drives_clone.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_drives_5236b9ee_f735_42fd_a236_17558f9e12d3_action_clone(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("drives_clone.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_drives_b02311e2_a83c_4c12_af10_b30d51c86913_action_resize(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("drives_resize.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_drives_9d1d2cf3_08c1_462f_8485_f4b073560809_action_resize(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("drives_resize.json")
        return (httplib.ACCEPTED, body, {}, httplib.responses[httplib.ACCEPTED])

    def _api_2_0_fwpolicies_detail(self, method, url, body, headers):
        body = self.fixtures.load("fwpolicies_detail.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_fwpolicies_CREATE_NO_RULES(self, method, url, body, headers):
        body = self.fixtures.load("fwpolicies_create_no_rules.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_fwpolicies_CREATE_WITH_RULES(self, method, url, body, headers):
        body = self.fixtures.load("fwpolicies_create_with_rules.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_servers_9de75ed6_fd33_45e2_963f_d405f31fd911_ATTACH_POLICY(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("servers_attach_policy.json")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_fwpolicies_0e339282_0cb5_41ac_a9db_727fb62ff2dc(self, method, url, body, headers):
        if method == "DELETE":
            body = ""
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _api_2_0_tags_detail(self, method, url, body, headers):
        body = self.fixtures.load("tags_detail.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_tags(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("tags_create.json")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_tags_WITH_RESOURCES(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("tags_create_with_resources.json")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_tags_a010ec41_2ead_4630_a1d0_237fa77e4d4d(self, method, url, body, headers):
        if method == "GET":
            # ex_get_tag
            body = self.fixtures.load("tags_get.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "PUT":
            # ex_tag_resource
            body = self.fixtures.load("tags_update.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "DELETE":
            # ex_delete_tag
            body = ""
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _api_2_0_balance(self, method, url, body, headers):
        body = self.fixtures.load("balance.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_pricing(self, method, url, body, headers):
        body = self.fixtures.load("pricing.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_currentusage(self, method, url, body, headers):
        body = self.fixtures.load("currentusage.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_subscriptions(self, method, url, body, headers):
        body = self.fixtures.load("subscriptions.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_subscriptions_STATUS_FILTER(self, method, url, body, headers):
        self.assertUrlContainsQueryParams(url, {"status": "active"})

        body = self.fixtures.load("subscriptions.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_subscriptions_RESOURCE_FILTER(self, method, url, body, headers):
        expected_params = {"resource": "cpu,mem", "status": "all"}
        self.assertUrlContainsQueryParams(url, expected_params)

        body = self.fixtures.load("subscriptions.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_subscriptions_7272_action_auto_renew(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_subscriptions_CREATE_SUBSCRIPTION(self, method, url, body, headers):
        body = self.fixtures.load("create_subscription.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_capabilities(self, method, url, body, headers):
        body = self.fixtures.load("capabilities.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_servers_availability_groups(self, method, url, body, headers):
        body = self.fixtures.load("servers_avail_groups.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_drives_availability_groups(self, method, url, body, headers):
        body = self.fixtures.load("drives_avail_groups.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_2_0_keypairs(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("keypairs_list.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "POST":
            body = self.fixtures.load("keypairs_import.json")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.CREATED])

    def _api_2_0_keypairs_186106ac_afb5_40e5_a0de_6f0feba5a3d5(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("keypairs_get.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "DELETE":
            body = ""
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import hashlib
import os.path
from io import FileIO as file

from libcloud.utils.py3 import b
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ContainerIsNotEmptyError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)


class DummyFileObject(file):
    def __init__(self, yield_count=5, chunk_len=10):
        self._yield_count = yield_count
        self._chunk_len = chunk_len

    def read(self, size):
        i = 0

        while i < self._yield_count:
            yield self._get_chunk(self._chunk_len)
            i += 1

    def _get_chunk(self, chunk_len):
        chunk = [str(x) for x in random.randint(97, 120)]
        return chunk

    def __len__(self):
        return self._yield_count * self._chunk_len


class DummyIterator:
    def __init__(self, data=None):
        self.hash = hashlib.md5()
        self._data = data or []
        self._current_item = 0

    def get_md5_hash(self):
        return self.hash.hexdigest()

    def next(self):
        if self._current_item == len(self._data):
            raise StopIteration

        value = self._data[self._current_item]
        self.hash.update(b(value))
        self._current_item += 1
        return value

    def __next__(self):
        return self.next()

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        pass


class DummyStorageDriver(StorageDriver):
    """
    Dummy Storage driver.

    >>> from libcloud.storage.drivers.dummy import DummyStorageDriver
    >>> driver = DummyStorageDriver('key', 'secret')
    >>> container = driver.create_container(container_name='test container')
    >>> container
    <Container: name=test container, provider=Dummy Storage Provider>
    >>> container.name
    'test container'
    >>> container.extra['object_count']
    0
    """

    name = "Dummy Storage Provider"
    website = "http://example.com"

    def __init__(self, api_key, api_secret):
        """
        :param    api_key:    API key or username to used (required)
        :type     api_key:    ``str``
        :param    api_secret: Secret password to be used (required)
        :type     api_secret: ``str``
        :rtype: ``None``
        """
        self._containers = {}

    def get_meta_data(self):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> driver.get_meta_data()['object_count']
        0
        >>> driver.get_meta_data()['container_count']
        0
        >>> driver.get_meta_data()['bytes_used']
        0
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container_name = 'test container 2'
        >>> container = driver.create_container(container_name=container_name)
        >>> obj = container.upload_object_via_stream(
        ...  object_name='test object', iterator=DummyFileObject(5, 10),
        ...  extra={})
        >>> driver.get_meta_data()['object_count']
        1
        >>> driver.get_meta_data()['container_count']
        2
        >>> driver.get_meta_data()['bytes_used']
        50

        :rtype: ``dict``
        """

        container_count = len(self._containers)
        object_count = sum(
            len(self._containers[container]["objects"]) for container in self._containers
        )

        bytes_used = 0
        for container in self._containers:
            objects = self._containers[container]["objects"]
            for _, obj in objects.items():
                bytes_used += obj.size

        return {
            "container_count": int(container_count),
            "object_count": int(object_count),
            "bytes_used": int(bytes_used),
        }

    def iterate_containers(self):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> list(driver.iterate_containers())
        []
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> container.name
        'test container 1'
        >>> container_name = 'test container 2'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 2, provider=Dummy Storage Provider>
        >>> container = driver.create_container(
        ...  container_name='test container 2')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerAlreadyExistsError:
        >>> container_list=list(driver.iterate_containers())
        >>> sorted([c.name for c in container_list])
        ['test container 1', 'test container 2']

        @inherits: :class:`StorageDriver.iterate_containers`
        """

        for container in list(self._containers.values()):
            yield container["container"]

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        container = self.get_container(container.name)
        objects = self._containers[container.name]["objects"].values()
        return self._filter_listed_container_objects(objects, prefix)

    def get_container(self, container_name):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> driver.get_container('unknown') #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerDoesNotExistError:
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> container.name
        'test container 1'
        >>> driver.get_container('test container 1')
        <Container: name=test container 1, provider=Dummy Storage Provider>

        @inherits: :class:`StorageDriver.get_container`
        """

        if container_name not in self._containers:
            raise ContainerDoesNotExistError(driver=self, value=None, container_name=container_name)

        return self._containers[container_name]["container"]

    def get_container_cdn_url(self, container):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> driver.get_container('unknown') #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerDoesNotExistError:
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> container.name
        'test container 1'
        >>> container.get_cdn_url()
        'http://www.test.com/container/test_container_1'

        @inherits: :class:`StorageDriver.get_container_cdn_url`
        """

        if container.name not in self._containers:
            raise ContainerDoesNotExistError(driver=self, value=None, container_name=container.name)

        return self._containers[container.name]["cdn_url"]

    def get_object(self, container_name, object_name):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> driver.get_object('unknown', 'unknown')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerDoesNotExistError:
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> driver.get_object(
        ...  'test container 1', 'unknown') #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ObjectDoesNotExistError:
        >>> obj = container.upload_object_via_stream(object_name='test object',
        ...      iterator=DummyFileObject(5, 10), extra={})
        >>> obj.name
        'test object'
        >>> obj.size
        50

        @inherits: :class:`StorageDriver.get_object`
        """

        self.get_container(container_name)
        container_objects = self._containers[container_name]["objects"]
        if object_name not in container_objects:
            raise ObjectDoesNotExistError(object_name=object_name, value=None, driver=self)

        return container_objects[object_name]

    def get_object_cdn_url(self, obj):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> obj = container.upload_object_via_stream(
        ...      object_name='test object 5',
        ...      iterator=DummyFileObject(5, 10), extra={})
        >>> obj.name
        'test object 5'
        >>> obj.get_cdn_url()
        'http://www.test.com/object/test_object_5'

        @inherits: :class:`StorageDriver.get_object_cdn_url`
        """

        container_name = obj.container.name
        container_objects = self._containers[container_name]["objects"]
        if obj.name not in container_objects:
            raise ObjectDoesNotExistError(object_name=obj.name, value=None, driver=self)

        return container_objects[obj.name].meta_data["cdn_url"]

    def create_container(self, container_name):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container
        <Container: name=test container 1, provider=Dummy Storage Provider>
        >>> container = driver.create_container(
        ...    container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerAlreadyExistsError:

        @inherits: :class:`StorageDriver.create_container`
        """

        if container_name in self._containers:
            raise ContainerAlreadyExistsError(
                container_name=container_name, value=None, driver=self
            )

        extra = {"object_count": 0}
        container = Container(name=container_name, extra=extra, driver=self)

        self._containers[container_name] = {
            "container": container,
            "objects": {},
            "cdn_url": "http://www.test.com/container/%s" % (container_name.replace(" ", "_")),
        }
        return container

    def delete_container(self, container):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container = Container(name = 'test container',
        ...    extra={'object_count': 0}, driver=driver)
        >>> driver.delete_container(container=container)
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerDoesNotExistError:
        >>> container = driver.create_container(
        ...      container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        >>> len(driver._containers)
        1
        >>> driver.delete_container(container=container)
        True
        >>> len(driver._containers)
        0
        >>> container = driver.create_container(
        ...    container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        >>> obj = container.upload_object_via_stream(
        ...   object_name='test object', iterator=DummyFileObject(5, 10),
        ...   extra={})
        >>> driver.delete_container(container=container)
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ContainerIsNotEmptyError:

        @inherits: :class:`StorageDriver.delete_container`
        """

        container_name = container.name
        if container_name not in self._containers:
            raise ContainerDoesNotExistError(container_name=container_name, value=None, driver=self)

        container = self._containers[container_name]
        if len(container["objects"]) > 0:
            raise ContainerIsNotEmptyError(container_name=container_name, value=None, driver=self)

        del self._containers[container_name]
        return True

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        kwargs_dict = {
            "obj": obj,
            "response": DummyFileObject(),
            "destination_path": destination_path,
            "overwrite_existing": overwrite_existing,
            "delete_on_failure": delete_on_failure,
        }

        return self._save_object(**kwargs_dict)

    def download_object_as_stream(self, obj, chunk_size=None):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container = driver.create_container(
        ...   container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        >>> obj = container.upload_object_via_stream(object_name='test object',
        ...    iterator=DummyFileObject(5, 10), extra={})
        >>> stream = container.download_object_as_stream(obj)
        >>> stream #doctest: +ELLIPSIS
        <...closed...>

        @inherits: :class:`StorageDriver.download_object_as_stream`
        """

        return DummyFileObject()

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
    ):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container_name = 'test container 1'
        >>> container = driver.create_container(container_name=container_name)
        >>> container.upload_object(file_path='/tmp/inexistent.file',
        ...     object_name='test') #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        LibcloudError:
        >>> file_path = path = os.path.abspath(__file__)
        >>> file_size = os.path.getsize(file_path)
        >>> obj = container.upload_object(file_path=file_path,
        ...                               object_name='test')
        >>> obj #doctest: +ELLIPSIS
        <Object: name=test, size=...>
        >>> obj.size == file_size
        True

        @inherits: :class:`StorageDriver.upload_object`
        """

        if not os.path.exists(file_path):
            raise LibcloudError(value="File %s does not exist" % (file_path), driver=self)

        size = os.path.getsize(file_path)
        return self._add_object(
            container=container, object_name=object_name, size=size, extra=extra
        )

    def upload_object_via_stream(self, iterator, container, object_name, extra=None, headers=None):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container = driver.create_container(
        ...    container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        >>> obj = container.upload_object_via_stream(
        ...   object_name='test object', iterator=DummyFileObject(5, 10),
        ...   extra={})
        >>> obj #doctest: +ELLIPSIS
        <Object: name=test object, size=50, ...>

        @inherits: :class:`StorageDriver.upload_object_via_stream`
        """

        size = len(iterator)
        return self._add_object(
            container=container, object_name=object_name, size=size, extra=extra
        )

    def delete_object(self, obj):
        """
        >>> driver = DummyStorageDriver('key', 'secret')
        >>> container = driver.create_container(
        ...   container_name='test container 1')
        ... #doctest: +IGNORE_EXCEPTION_DETAIL
        >>> obj = container.upload_object_via_stream(object_name='test object',
        ...   iterator=DummyFileObject(5, 10), extra={})
        >>> obj #doctest: +ELLIPSIS
        <Object: name=test object, size=50, ...>
        >>> container.delete_object(obj=obj)
        True
        >>> obj = Object(name='test object 2',
        ...    size=1000, hash=None, extra=None,
        ...    meta_data=None, container=container,driver=None)
        >>> container.delete_object(obj=obj) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ObjectDoesNotExistError:

        @inherits: :class:`StorageDriver.delete_object`
        """

        container_name = obj.container.name
        object_name = obj.name
        obj = self.get_object(container_name=container_name, object_name=object_name)

        del self._containers[container_name]["objects"][object_name]
        return True

    def _add_object(self, container, object_name, size, extra=None):
        container = self.get_container(container.name)

        extra = extra or {}
        meta_data = extra.get("meta_data", {})
        meta_data.update(
            {"cdn_url": "http://www.test.com/object/%s" % (object_name.replace(" ", "_"))}
        )
        obj = Object(
            name=object_name,
            size=size,
            extra=extra,
            hash=None,
            meta_data=meta_data,
            container=container,
            driver=self,
        )

        self._containers[container.name]["objects"][object_name] = obj
        return obj


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

import requests
import requests_mock

from libcloud.http import LibcloudConnection
from libcloud.common.base import Connection, XmlResponse, JsonResponse
from libcloud.common.types import MalformedResponseError


class ResponseClassesTests(unittest.TestCase):
    def setUp(self):
        self.mock_connection = LibcloudConnection(host="mock.com", port=80)
        self.mock_connection.driver = None

    def test_XmlResponse_class(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/2", text="<foo>bar</foo>")
            response_obj = requests.get("mock://test.com/2")
            response = XmlResponse(response=response_obj, connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed.tag, "foo")
        self.assertEqual(parsed.text, "bar")

    def test_XmlResponse_class_malformed_response(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/", text="<foo>")
            response_obj = requests.get("mock://test.com/")
            try:
                XmlResponse(response=response_obj, connection=self.mock_connection)
            except MalformedResponseError:
                pass
            else:
                self.fail("Exception was not thrown")

    def test_XmlResponse_class_zero_length_body_strip(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/", text=" ")
            response_obj = requests.get("mock://test.com/")
            response = XmlResponse(response=response_obj, connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, "")

    def test_JsonResponse_class_success(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/", text='{"foo": "bar"}')
            response_obj = requests.get("mock://test.com/")
            response = JsonResponse(response=response_obj, connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, {"foo": "bar"})

    def test_JsonResponse_class_malformed_response(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/", text='{"foo": "bar"')
            response_obj = requests.get("mock://test.com/")
            try:
                JsonResponse(response=response_obj, connection=self.mock_connection)
            except MalformedResponseError:
                pass
            else:
                self.fail("Exception was not thrown")

    def test_JsonResponse_class_zero_length_body_strip(self):
        with requests_mock.mock() as m:
            m.register_uri("GET", "mock://test.com/", text=" ")
            response_obj = requests.get("mock://test.com/")
            response = JsonResponse(response=response_obj, connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, "")

    def test_RawResponse_class_read_method(self):
        """
        Test that the RawResponse class includes a response
        property which exhibits the same properties and methods
        as httplib.HTTPResponse for backward compat <1.5.0
        """
        TEST_DATA = "1234abcd"

        conn = Connection(host="mock.com", port=80, secure=False)
        conn.connect()

        with requests_mock.Mocker() as m:
            m.register_uri(
                "GET",
                "http://mock.com/raw_data",
                text=TEST_DATA,
                headers={"test": "value"},
            )
            response = conn.request("/raw_data", raw=True)
        data = response.response.read()
        self.assertEqual(data, TEST_DATA)

        header_value = response.response.getheader("test")
        self.assertEqual(header_value, "value")

        headers = response.response.getheaders()
        self.assertEqual(headers, [("test", "value")])

        self.assertEqual(response.response.status, 200)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.dns.types import RecordType
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DIGITALOCEAN_v2_PARAMS
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.digitalocean import DigitalOceanDNSDriver


class DigitalOceanDNSTests(LibcloudTestCase):
    def setUp(self):
        DigitalOceanDNSDriver.connectionCls.conn_class = DigitalOceanDNSMockHttp
        DigitalOceanDNSMockHttp.type = None
        self.driver = DigitalOceanDNSDriver(*DIGITALOCEAN_v2_PARAMS)

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertTrue(len(zones) >= 1)

    def test_get_zone(self):
        zone = self.driver.get_zone("testdomain")
        self.assertEqual(zone.id, "testdomain")

    def test_get_zone_not_found(self):
        DigitalOceanDNSMockHttp.type = "NOT_FOUND"
        self.assertRaises(Exception, self.driver.get_zone, "testdomain")

    def test_list_records(self):
        zone = self.driver.get_zone("testdomain")
        records = self.driver.list_records(zone)
        self.assertTrue(len(records) >= 1)
        self.assertEqual(records[1].ttl, 1800)
        self.assertEqual(records[4].ttl, None)

    def test_get_record(self):
        record = self.driver.get_record("testdomain", "1234564")
        self.assertEqual(record.id, "1234564")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "123.45.67.89")
        self.assertEqual(record.ttl, 1800)

    def test_get_record_not_found(self):
        DigitalOceanDNSMockHttp.type = "NOT_FOUND"
        self.assertRaises(Exception, self.driver.get_zone, "testdomain")

    def test_create_zone(self):
        DigitalOceanDNSMockHttp.type = "CREATE"
        zone = self.driver.create_zone("testdomain")
        self.assertEqual(zone.id, "testdomain")

    def test_create_record(self):
        zone = self.driver.get_zone("testdomain")

        DigitalOceanDNSMockHttp.type = "CREATE"
        record = self.driver.create_record(
            "sub", zone, RecordType.A, "234.56.78.90", extra={"ttl": 60}
        )
        self.assertEqual(record.id, "1234565")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "234.56.78.90")
        self.assertEqual(record.ttl, 60)

    def test_update_record(self):
        record = self.driver.get_record("testdomain", "1234564")

        DigitalOceanDNSMockHttp.type = "UPDATE"
        record = self.driver.update_record(record, data="234.56.78.90", extra={"ttl": 60})
        self.assertEqual(record.id, "1234564")
        self.assertEqual(record.data, "234.56.78.90")
        self.assertEqual(record.ttl, 60)

    def test_delete_zone(self):
        zone = self.driver.get_zone("testdomain")

        DigitalOceanDNSMockHttp.type = "DELETE"
        self.assertTrue(self.driver.delete_zone(zone))

    def test_delete_record(self):
        record = self.driver.get_record("testdomain", "1234564")

        DigitalOceanDNSMockHttp.type = "DELETE"
        self.assertTrue(self.driver.delete_record(record))


class DigitalOceanDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures("digitalocean")

    response_map = {
        None: httplib.OK,
        "CREATE": httplib.CREATED,
        "DELETE": httplib.NO_CONTENT,
        "EMPTY": httplib.OK,
        "NOT_FOUND": httplib.NOT_FOUND,
        "UNAUTHORIZED": httplib.UNAUTHORIZED,
        "UPDATE": httplib.OK,
        "UNPROCESSABLE": httplib.UNPROCESSABLE_ENTITY,
    }

    def _v2_domains(self, method, url, body, headers):
        body = self.fixtures.load("_v2_domains.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_CREATE(self, method, url, body, headers):
        if body is None:
            body = self.fixtures.load("_v2_domains_UNPROCESSABLE_ENTITY.json")
            return (
                self.response_map["UNPROCESSABLE"],
                body,
                {},
                httplib.responses[self.response_map["UNPROCESSABLE"]],
            )
        body = self.fixtures.load("_v2_domains_CREATE.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain(self, method, url, body, headers):
        body = self.fixtures.load("_v2_domains_testdomain.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_DELETE(self, method, url, body, headers):
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_NOT_FOUND(self, method, url, body, headers):
        body = self.fixtures.load("_v2_domains_testdomain_NOT_FOUND.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_records(self, method, url, body, headers):
        body = self.fixtures.load("_v2_domains_testdomain_records.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_records_CREATE(self, method, url, body, headers):
        if body is None:
            body = self.fixtures.load("_v2_domains_UNPROCESSABLE_ENTITY.json")
            return (
                self.response_map["UNPROCESSABLE"],
                body,
                {},
                httplib.responses[self.response_map["UNPROCESSABLE"]],
            )
        body = self.fixtures.load("_v2_domains_testdomain_records_CREATE.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_records_1234564(self, method, url, body, headers):
        body = self.fixtures.load("_v2_domains_testdomain_records_1234564.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_records_1234564_DELETE(self, method, url, body, headers):
        self.type = "DELETE"
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )

    def _v2_domains_testdomain_records_1234564_UPDATE(self, method, url, body, headers):
        if body is None:
            body = self.fixtures.load("_v2_domains_UNPROCESSABLE_ENTITY.json")
            return (
                self.response_map["UNPROCESSABLE"],
                body,
                {},
                httplib.responses[self.response_map["UNPROCESSABLE"]],
            )
        body = self.fixtures.load("_v2_domains_testdomain_records_1234564_UPDATE.json")
        return (
            self.response_map[self.type],
            body,
            {},
            httplib.responses[self.response_map[self.type]],
        )


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from libcloud.common.base import Response, ConnectionUserAndKey
from libcloud.common.types import ProviderError

OK_CODES = ["200", "211", "212", "213"]
ERROR_CODES = [
    "401",
    "403",
    "405",
    "406",
    "407",
    "408",
    "409",
    "410",
    "411",
    "412",
    "413",
    "414",
    "450",
    "451",
]


class WorldWideDNSException(ProviderError):
    def __init__(self, value, http_code, code, driver=None):
        self.code = code
        super().__init__(value, http_code, driver)


class SuspendedAccount(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Login ID you supplied is SUSPENDED, you need to renew" + " your account"
        super().__init__(value, http_code, 401, driver)


class LoginOrPasswordNotMatch(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Login ID and/or Password you supplied is not on file or" + " does not match"
        super().__init__(value, http_code, 403, driver)


class NonExistentDomain(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Domain name supplied is not in your account"
        super().__init__(value, http_code, 405, driver)


class CouldntRemoveDomain(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Error occurred removing domain from name server, try again"
        super().__init__(value, http_code, 406, driver)


class LimitExceeded(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Your limit was exceeded, you need to upgrade your account"
        super().__init__(value, http_code, 407, driver)


class ExistentDomain(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Domain already exists on our servers"
        super().__init__(value, http_code, 408, driver)


class DomainBanned(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Domain is listed in DNSBL and is banned from our servers"
        super().__init__(value, http_code, 409, driver)


class InvalidDomainName(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Invalid domain name"
        super().__init__(value, http_code, 410, driver)


class ErrorOnReloadInNameServer(WorldWideDNSException):
    def __init__(self, server, http_code, driver=None):
        if server == 1:
            value = "Name server #1 kicked an error on reload, contact support"
            code = 411
        elif server == 2:
            value = "Name server #2 kicked an error on reload, contact support"
            code = 412
        elif server == 3:
            value = "Name server #3 kicked an error on reload, contact support"
            code = 413
        super().__init__(value, http_code, code, driver)


class NewUserNotValid(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "New userid is not valid"
        super().__init__(value, http_code, 414, driver)


class CouldntReachNameServer(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "Couldn't reach the name server, try again later"
        super().__init__(value, http_code, 450, driver)


class NoZoneFile(WorldWideDNSException):
    def __init__(self, http_code, driver=None):
        value = "No zone file in the name server queried"
        super().__init__(value, http_code, 451, driver)


ERROR_CODE_TO_EXCEPTION_CLS = {
    "401": SuspendedAccount,
    "403": LoginOrPasswordNotMatch,
    "405": NonExistentDomain,
    "406": CouldntRemoveDomain,
    "407": LimitExceeded,
    "408": ExistentDomain,
    "409": DomainBanned,
    "410": InvalidDomainName,
    "411": ErrorOnReloadInNameServer,
    "412": ErrorOnReloadInNameServer,
    "413": ErrorOnReloadInNameServer,
    "414": NewUserNotValid,
    "450": CouldntReachNameServer,
    "451": NoZoneFile,
}


class WorldWideDNSResponse(Response):
    def parse_body(self):
        """
        Parse response body.

        :return: Parsed body.
        :rtype: ``str``
        """
        if self._code_response(self.body):
            codes = re.split("\r?\n", self.body)
            for code in codes:
                if code in OK_CODES:
                    continue
                elif code in ERROR_CODES:
                    exception = ERROR_CODE_TO_EXCEPTION_CLS.get(code)
                    if code in ["411", "412", "413"]:
                        server = int(code[2])
                        raise exception(server, self.status)
                    raise exception(self.status)
        return self.body

    def _code_response(self, body):
        """
        Checks if the response body contains code status.

        :rtype: ``bool``
        """
        available_response_codes = OK_CODES + ERROR_CODES
        codes = re.split("\r?\n", body)
        if codes[0] in available_response_codes:
            return True
        return False


class WorldWideDNSConnection(ConnectionUserAndKey):
    host = "www.worldwidedns.net"
    responseCls = WorldWideDNSResponse

    def add_default_params(self, params):
        """
        Add parameters that are necessary for every request

        This method adds ``NAME`` and ``PASSWORD`` to
        the request.
        """
        params["NAME"] = self.user_id
        params["PASSWORD"] = self.key

        reseller_id = getattr(self, "reseller_id", None)
        if reseller_id:
            params["ID"] = reseller_id

        return params
